"""
rail module
===========
Classes:
"""

from .grid import GridTools
from .plottools import make_substructures as make_substruct
from .tools.utils import hypot_func
from . import constants as sfc
from . import units as sfu
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.axes import Axes
import numpy as np
from astropy import units as u
from scipy.interpolate import griddata
from collections.abc import Iterable
from skimage import measure
import numbers

get_sky_from_disc_coords = GridTools.get_sky_from_disc_coords

class Rail(object):
    def __init__(self, model, prop, coord_levels=None):
        """
        Initialise Rail class. This class provides useful methods to compute azimuthal/radial contours and/or azimuthal averages from an input ``prop`` map projected on the sky.
        The methods in this class use the model disc vertical structure to convert sky to disc coordinates, and therefore all quantities returned are referred to the disc reference frame.

        Parameters
        ----------
        model : `.disc2d.General2d` instance
           Model to get the disc vertical structure from.

        prop : array_like, shape (nx, ny)
           Observable to be analysed, as projected on the sky.
        
        coord_levels : array_like, shape (nlevels,), optional
           Radial or azimuthal levels where *prop* will be mapped.

              - If None, coord_levels are assumed to vary linearly and radially as np.arange(model.Rmin, model.Rmax, beam/4)
              - Does not have any effect on the `Rail.make_filaments` method
        """
        
        _rail_coords = model.projected_coords
        X, Y = model.skygrid['meshgrid']        
        self.X = X/sfu.au
        self.Y = Y/sfu.au

        self.R = _rail_coords['R']
        self.phi = _rail_coords['phi']
        self.R_nonan = _rail_coords['R_nonan']                
        self.extent = model.skygrid['extent']
        self.dpc = model.dpc
        self.Rmin = model.Rmin
        self.Rmax = model.Rmax        

        self.beam_size = model.beam_size
        self.header = model.header

        self.prop = prop        
        if coord_levels is None:
            self.coord_levels = np.arange(
                model.Rmin.to('au').value,
                model.Rmax.to('au').value,
                model.beam_size.to('au').value/4
            )
        else:
            self.coord_levels = coord_levels

        self._lev_list = None
        self._coord_list = None
        self._resid_list = None
        self._color_list = None

        self.model = model

    @staticmethod
    def make_contour_lev(prop, lev, X, Y, acc_threshold=20): 
        contour = measure.find_contours(prop, lev)
        ind_good = np.argmin([np.abs(lev-prop[tuple(np.round(contour[i][0]).astype(int))]) for i in range(len(contour))]) #get contour id closest to lev              
        inds_cont = np.round(contour[ind_good]).astype(int)
        inds_cont = [tuple(f) for f in inds_cont]
        first_cont = np.array([prop[i] for i in inds_cont])
        corr_inds = np.abs(first_cont-lev) < acc_threshold
        x_cont = np.array([X[i] for i in inds_cont])
        y_cont = np.array([Y[i] for i in inds_cont])
        return x_cont[corr_inds], y_cont[corr_inds], inds_cont, corr_inds
        
    @staticmethod
    def beams_along_annulus(lev, Rgrid, beam_size, X, Y):
        xc, yc, _, _ = Rail.make_contour_lev(Rgrid, lev, X, Y)
        try:
            rc = hypot_func(xc, yc)
            a = np.max(rc)
            b = np.min(rc)
            ellipse_perim = np.pi*(3*(a+b)-np.sqrt((3*a+b)*(a+3*b))) #Assumes elliptical path is not affected much by the disc vertical structure
            return ellipse_perim/beam_size
        except ValueError: #No contour found
            return np.inf
        
    def prop_along_coords(self,
                          coord_ref=None,
                          surface='upper',
                          ax=None,
                          ax2=None,
                          acc_threshold=10, #0.05
                          max_prop_threshold=np.inf,
                          color_bounds=[200, 400],
                          colors=['red', 'dodgerblue', '#FFB000'],
                          lws=[0.3, 0.3, 0.3], lw_ax2_factor=1,
                          fold=False,
                          fold_func=np.subtract):
        """
        Compute azimuthal contours according to the model disc geometry 
        to retrieve and plot information from the input two-dimensional map ``prop``.    

        Parameters
        ----------
        coord_ref : scalar, optional
           The contour at this coordinate will be highlighted in bold black.

        ax : `matplotlib.axes` instance, optional
           ax instance where computed contours are to be shown
                   
        ax2 : `matplotlib.axes` instance (or list of instances), optional
           Additional ax(s) instance(s) to plot the location of contours in the disc. 
           
        acc_threshold : float, optional 
           Threshold to accept points on contours at a given coord_level. If coord_level obtained for a pixel is such that np.abs(level_pixel-level_reference)<acc_threshold the pixel value is accepted

        max_prop_threshold : float, optional 
           Threshold to accept points of contours. Rejects residuals of the contour if they are >= max_prop_threshold. Useful to reject hot pixels.

        color_bounds : array_like, shape (nbounds,), optional
           Color bounds for contour colors.
           
        colors : array_like, shape (nbounds+1,), optional
           Contour colors within bounds. 

        lws : array_like, shape (nbounds+1), optional
           Contour linewidths within bounds.

        lw_ax2_factor : float, optional
           Fraction of lws for linewidth value of contours in ax2.

        fold : bool, optional
           If True, subtract residuals by folding along the projected minor axis of the disc. Currently working for azimuthal contours only.
           
        fold_func : function, optional
           If fold, this function is used to operate between folded quadrants. Defaults to np.subtract.
        """

        prop, coord_levels = self.prop, self.coord_levels
        _rail_phi = self.phi[surface] #np.where(self.phi['upper'] < 0.98*np.pi, self.phi['upper'], np.nan)
        coords_azimuthal = [self.R_nonan[surface]/sfu.au, np.degrees(_rail_phi)]
        coords_radial = [] #Missing this, must add the option to allow the user switch between azimuthal or radial contours
        
        X, Y = self.X, self.Y
        coords = coords_azimuthal

        if isinstance(color_bounds, Iterable):
            color_bounds = np.insert([0, np.inf], 1, color_bounds)
            nbounds = len(color_bounds)
        else:
            nbounds = 0
            
        coord_list, lev_list, resid_list, color_list = [], [], [], []

        if np.sum(coord_levels==coord_ref)==0 and coord_ref is not None:
            coord_levels = np.append(coord_levels, coord_ref)

        for levi, lev in enumerate(coord_levels):

            contour = measure.find_contours(coords[0], lev) #, fully_connected='high', positive_orientation='high')

            if len(contour)==0:
                print ('no contours found for phi =', lev)
                continue
            ind_good = np.argmin([np.abs(lev-coords[0][tuple(np.round(contour[i][0]).astype(int))]) for i in range(len(contour))]) #get contour id closest to lev
            inds_cont = np.round(contour[ind_good]).astype(int)
            inds_cont = [tuple(f) for f in inds_cont]
            first_cont = np.array([coords[0][i] for i in inds_cont])
            second_cont = np.array([coords[1][i] for i in inds_cont])
            prop_cont = np.array([prop[i] for i in inds_cont])

            corr_inds = np.abs(first_cont-lev) < acc_threshold #clean based on acc_threshold
            
            _, tmp = np.unique(second_cont, return_index=True) 
            unique_inds = np.zeros_like(second_cont).astype(bool)
            unique_inds[tmp] = True
            corr_inds = corr_inds & unique_inds #make sure points are not repeated

            if lev == coord_ref: zorder=10
            else: zorder=np.random.randint(0,10)

            if color_bounds is None:
                if isinstance(colors, Iterable) and not isinstance(colors, str):
                    color = colors[levi]
                else:
                    color = colors
                    
                if isinstance(lws, Iterable):
                    lw = lws[levi]
                else:
                    lw = lws
            else:
                for i,bound in enumerate(color_bounds):
                    if lev == coord_ref: 
                        lw = 2.0
                        color = 'k'
                        zorder = 10
                        break
                    elif i!=(nbounds-1):
                        if color_bounds[i] <= lev <= color_bounds[i+1]:
                            lw = lws[i]
                            color = colors[i]                                            
                            break
                    else:
                        lw = lws[i]
                        color = colors[i]                    
                        break
                            
            if fold:
                #if lev < color_bounds[0]: continue
                ref_pos = 90 #Reference axis for positive angles
                ref_neg = -90
                angles = second_cont[corr_inds]
                prop_ = prop_cont[corr_inds]
                angles_pos = angles[angles>=0] #0, 180
                angles_neg = angles[angles<0] #-180, 0
                relative_diff_pos = ref_pos - angles_pos
                relative_diff_neg = ref_neg - angles_neg
                angle_diff_pos, prop_diff_pos = [], []
                angle_diff_neg, prop_diff_neg = [], []

                for i,diff in enumerate(relative_diff_pos):
                    #Finding where the difference matches that of the current analysis angle
                    #The -1 flips the sign so that the number on the other side of the symmetry axis is found                
                    ind = np.argmin(np.abs(-1*relative_diff_pos - diff))  
                    mirror_ind = angles==angles_pos[ind]
                    current_ind = angles==angles_pos[i]
                    prop_diff = fold_func(prop_[current_ind][0], prop_[mirror_ind][0])
                    angle_diff_pos.append(angles_pos[i])
                    prop_diff_pos.append(prop_diff)
                angle_diff_pos = np.asarray(angle_diff_pos)
                prop_diff_pos = np.asarray(prop_diff_pos)

                if len(angle_diff_pos)>1:
                    ind_sort_pos = np.argsort(angle_diff_pos)
                    plot_ang_diff_pos = angle_diff_pos[ind_sort_pos]
                    plot_prop_diff_pos = prop_diff_pos[ind_sort_pos]
                    ind_prop_pos = np.abs(plot_prop_diff_pos)<max_prop_threshold
                    if ax is not None:
                        ax.plot(plot_ang_diff_pos[ind_prop_pos], plot_prop_diff_pos[ind_prop_pos],
                                color=color, lw=lw, zorder=zorder)
                    coord_list.append(plot_ang_diff_pos[ind_prop_pos])
                    resid_list.append(plot_prop_diff_pos[ind_prop_pos])
                    color_list.append(color)
                    lev_list.append(lev)
                else: 
                    plot_ang_diff_pos = []
                    plot_prop_diff_pos = []

                for i,diff in enumerate(relative_diff_neg):
                    ind = np.argmin(np.abs(-1*relative_diff_neg - diff))
                    mirror_ind = angles==angles_neg[ind]
                    current_ind = angles==angles_neg[i]
                    prop_diff = fold_func(prop_[current_ind][0], prop_[mirror_ind][0])
                    angle_diff_neg.append(angles_neg[i])
                    prop_diff_neg.append(prop_diff)
                angle_diff_neg = np.asarray(angle_diff_neg)
                prop_diff_neg = np.asarray(prop_diff_neg)

                if len(angle_diff_neg)>1:
                    ind_sort_neg = np.argsort(np.abs(angle_diff_neg))
                    plot_ang_diff_neg = angle_diff_neg[ind_sort_neg]
                    plot_prop_diff_neg = prop_diff_neg[ind_sort_neg]
                    ind_prop_neg = np.abs(plot_prop_diff_neg)<max_prop_threshold
                    if ax is not None:
                        ax.plot(plot_ang_diff_neg[ind_prop_neg], plot_prop_diff_neg[ind_prop_neg],
                                color=color, lw=lw, zorder=zorder)
                    coord_list.append(plot_ang_diff_neg[ind_prop_neg])
                    resid_list.append(plot_prop_diff_neg[ind_prop_neg])
                    color_list.append(color)
                    lev_list.append(lev)
                else: 
                    plot_ang_diff_neg = []
                    plot_prop_diff_neg = []

                """
                if len(angle_diff_pos)>1 or len(angle_diff_neg)>1:
                    coord_list.append(np.append(plot_ang_diff_pos, plot_ang_diff_neg))
                    resid_list.append(np.append(plot_prop_diff_pos, plot_prop_diff_neg))
                    color_list.append(color)
                    lev_list.append(lev)
                """
            else:
                ind_sort = np.argsort(second_cont[corr_inds]) #sorting by azimuth/radius to avoid 'joint' boundaries in plot
                coord_list.append(second_cont[corr_inds][ind_sort])
                resid_list.append(prop_cont[corr_inds][ind_sort])
                color_list.append(color)
                lev_list.append(lev)

                if ax is not None:
                    ax.plot(second_cont[corr_inds][ind_sort], prop_cont[corr_inds][ind_sort], color=color, lw=lw, zorder=zorder)

            if ax2 is not None:
                x_cont = np.array([X[i] for i in inds_cont])
                y_cont = np.array([Y[i] for i in inds_cont])
            if isinstance(ax2, Axes):
                ax2.plot(x_cont[corr_inds], y_cont[corr_inds], color=color, lw=lw*lw_ax2_factor)
            elif isinstance(ax2, list):
                for axi in ax2: 
                    if isinstance(axi, Axes):
                        axi.plot(x_cont[corr_inds], y_cont[corr_inds], color=color, lw=lw*lw_ax2_factor)

        self._lev_list = lev_list
        self._coord_list = coord_list
        self._resid_list = resid_list
        self._color_list = color_list

        return [np.asarray(tmp, dtype=dty)
                for tmp,dty in zip([lev_list, coord_list, resid_list, color_list],
                                   [float, object, object, object]
                )
        ]

    
    def get_average(self, surface='upper', 
                    av_func=np.nanmean,
                    mask_ang=0,
                    sigma_thres=np.inf,
                    mask_from_map=None, mask_perc_init=0.2,
                    plot_diagnostics=False, tag='',
                    forward_error=False, error_func=False, error_unit=1.0, error_thres=np.inf,                    
                    **kwargs_along_coords):
        #mask_ang: +- angles to reject around minor axis (i.e. phi=+-90) 

        if self._lev_list is None:
            kwargs_along_coords.update({'surface': surface})
            self.prop_along_coords(**kwargs_along_coords)

        lev_list, coord_list, resid_list = self._lev_list, self._coord_list, self._resid_list
        Rgrid = self.R_nonan[surface]/sfu.au
        X = self.X
        Y = self.Y
        beam_size = self.beam_size.to('au').value
        
        frac_annulus = 1.0 #if halves, 0.5; if quadrants, 0.25
        nconts = len(lev_list)

        if mask_from_map is not None:
            mrail = Rail(self.model, mask_from_map, lev_list)
            _, coord_list2, resid_list2, _ = mrail.prop_along_coords(surface=surface)
            
            resid_thres = []
            ind_accep = []
            mean_list, sigma_list = [], []
            
            for i in range(nconts):
                resid2_nonan = resid_list2[i][~np.isnan(resid_list2[i])]
                psig = int(mask_perc_init*len(resid2_nonan))
                isort = np.argsort(resid2_nonan)
                sigma = np.nanstd(resid2_nonan[isort][psig:-psig])
                mean_val = np.nanmean(resid2_nonan[isort][psig:-psig])
                ind = np.abs(resid_list2[i]-mean_val)<sigma_thres*sigma

                ind_accep.append(
                    (
                        ((coord_list[i]<90-mask_ang) & (coord_list[i]>-90+mask_ang)) |
                        ((coord_list[i]>90+mask_ang) | (coord_list[i]<-90-mask_ang))
                    )
                    & ind
                    )
                mean_list.append(mean_val)
                sigma_list.append(sigma)
                
            if plot_diagnostics:
                fig, ax = plt.subplots(nrows=2, figsize=(12,10))                
                ax0, ax1 = ax
                idiag = (np.array([0.3, 0.6, 0.9])*nconts).astype(int) #ind of radii to be plotted

                for i in idiag:
                    ax0.plot(coord_list2[i], resid_list2[i], lw=2.5, alpha=0.5, label='R=%.1f'%lev_list[i])
                    ax0.scatter(coord_list2[i][ind_accep[i]], resid_list2[i][ind_accep[i]], ec='k', fc='none', s=40, lw=1.2)
                    ax1.plot(coord_list[i], resid_list[i], lw=2.5, alpha=0.5)
                    ax1.scatter(coord_list[i][ind_accep[i]], resid_list[i][ind_accep[i]], ec='k', fc='none', s=40, lw=1.2)

                i = idiag[0]
                yfac = 10                
                ax0.set_ylim(mean_list[idiag[-1]]-yfac*sigma_list[idiag[-1]], mean_list[i]+yfac*sigma_list[i])
                mean_eval_map = np.nanmean(resid_list[i][ind_accep[i]])
                sigma_eval_map = np.nanstd(resid_list[i][ind_accep[i]])
                ax1.set_ylim(mean_eval_map-yfac*sigma_eval_map, mean_eval_map+yfac*sigma_eval_map)

                tick_angles = np.arange(-150, 180, 30)
                ax0.legend(frameon=False, fontsize=15)

                for axi in ax:
                    axi.set_xticks(tick_angles)
                    axi.set_xlabel(r'Azimuth [deg]')                        

                fig.savefig('diagnostics_mask_average_from_map_%s.png'%tag)
                plt.close()
                
        else:
            #anything higher than sigma_thres will be rejected from annulus
            resid_thres = [sigma_thres*np.nanstd(resid_list[i]) for i in range(nconts)] 

            ind_accep = [(((coord_list[i]<90-mask_ang) & (coord_list[i]>-90+mask_ang)) |
                          ((coord_list[i]>90+mask_ang) | (coord_list[i]<-90-mask_ang))) &
                         (np.abs(resid_list[i]-np.nanmean(resid_list[i]))<resid_thres[i])
                         for i in range(nconts)]

        av_annulus = np.array([av_func(resid_list[i][ind_accep[i]]) for i in range(nconts)])

        if error_func is None:
            av_error = None
            
        else:
            beams_ring_sqrt = np.sqrt([frac_annulus*Rail.beams_along_annulus(lev, Rgrid, beam_size, X, Y) for lev in lev_list])
            if callable(error_func): #if error map provided, compute average error per radius, divided by sqrt of number of beams (see Michiel Hogerheijde notes on errors)
                av_error = np.zeros(nconts)
                for i in range(nconts):
                    x_accep, y_accep, __ = get_sky_from_disc_coords(lev_list[i], coord_list[i][ind_accep[i]]) #MISSING z, incl, PA for the function to work
                    error_accep = np.array(list(map(error_func, x_accep, y_accep))).T[0]
                    sigma2_accep = np.where((np.isfinite(error_accep)) & (error_unit*error_accep<error_thres) & (error_accep>0), (error_unit*error_accep)**2, 0)
                    Np_accep = len(coord_list[i][ind_accep[i]])
                    av_error[i] = np.sqrt(np.nansum(sigma2_accep)/Np_accep)/beams_ring_sqrt[i]  
            else: #compute standard error of mean value
                if mask_from_map is not None and forward_error:
                    resid = resid_list2
                else:
                    resid = resid_list
                av_error = np.array([np.std(resid[i][ind_accep[i]], ddof=1) for i in range(nconts)])/beams_ring_sqrt

        return av_annulus, av_error


    def get_average_zones(self, surface='upper',
                          av_func=np.nanmean,
                          az_zones=[[-30, 30], [150,  -150]],
                          fast=True, #If not fast use np.trapz integral instead of av_func to get average values
                          join_zones=False, #Get a single averaged profile from all values in the input zones
                          sigma_thres=np.inf, error_func=False, error_unit=1.0, error_thres=np.inf,
                          **kwargs_along_coords):
                          
        if self._lev_list is None:
            kwargs_along_coords.update({'surface': surface})
            self.prop_along_coords(**kwargs_along_coords)

        lev_list, coord_list, resid_list = self._lev_list, self._coord_list, self._resid_list
        Rgrid = self.R_nonan[surface]/sfu.au
        X = self.X
        Y = self.Y
        beam_size = self.beam_size.to('au').value
        
        nconts = len(lev_list)
        nzones = len(az_zones)

        resid_thres = [sigma_thres*np.nanstd(resid_list[i]) for i in range(nconts)] 

        make_or = lambda az0, az1: [((coord_list[i]>az0) | (coord_list[i]<az1))
                                    & (np.abs(resid_list[i]-np.nanmean(resid_list[i])) < resid_thres[i])
                                    for i in range(nconts)]
        
        make_and = lambda az0, az1: [((coord_list[i]>az0) & (coord_list[i]<az1))
                                     & (np.abs(resid_list[i]-np.nanmean(resid_list[i])) < resid_thres[i])
                                     for i in range(nconts)]

        def get_portion_inds(az):
            az0, az1 = az
            if (az0 > az1):
                inds = make_or(az0, az1)
            else:
                inds = make_and(az0, az1)
            return inds

        def get_portion_percent(az):
            az0, az1 = az
            if (az0 > az1):
                if az0 < 0: perc = 1 - (az0-az1)/360.
                else: perc = (180-az0 + 180-np.abs(az1))/360.
            else:
                perc = (az1-az0)/360.
            return perc

        #inds containts lists of indices, one list per zone. Each is a list of lists, with as many lists as nconts (number of radii). Each sublist has different number of indices, the larger the radius (i.e. larger path) the more indices.        
        inds = [get_portion_inds(zone) for zone in az_zones] 
        az_percent = np.array([get_portion_percent(zone) for zone in az_zones])

        if join_zones and nzones>1:
            if not fast:
                warnings.warn('Using standard (fast) algorithm to compute average of values within zones altogether')

            #concatenate indices from zones, per radius
            inds = [[functools.reduce(lambda x,y: x+y, [ind[i] for ind in inds]) for i in range(nconts)]] 
            az_percent = np.atleast_1d(np.sum(az_percent))
            nzones = 1

        if fast: #compute 'arithmetic' average using av_func
            av_on_inds = [np.array([av_func(resid_list[i][ind[i]]) for i in range(nconts)]) for ind in inds]        

        else: #Compute average using integral definition (trapezoid seems to do better than simpson)
            av_integral = lambda y,x,dT: np.trapz(y, x=x)/dT # or trapezoid from scipy.integrate
            av_on_inds = []

            for ind in inds:

                av_annulus = []

                for i in range(nconts):
                    ii = ind[i]
                    coords_ii = coord_list[i][ii]
                    resid_ii = resid_list[i][ii]

                    if not len(coords_ii):
                        trap = None
                    else:
                        trap = av_integral(resid_ii, coords_ii, coords_ii[-1]-coords_ii[0]) #dT assumes coords_list is sorted (no matter if ascending or descending)
                    av_annulus.append(trap) 
                av_on_inds.append(np.array(av_annulus))
            
        beams_ring_full = [Rail.beams_along_annulus(lev, Rgrid, beam_size, X, Y) for lev in lev_list]
        beams_zone_sqrt = [np.sqrt(az_percent*br) for br in beams_ring_full]

        if error_func is None:
            av_error = None

        else:
            if callable(error_func): #Not yet tested
                #if error map provided, compute average error per radius, divided by sqrt of number of beams (see Michiel Hogerheijde notes on errors)  
                av_error = []
                for i in range(nconts):
                    r_ind = [get_sky_from_disc_coords(lev_list[i], coord_list[i][ind[i]]) for ind in inds] #MISSING z, incl, PA for the function to work
                    error_ind = [np.array(list(map(error_func, r_ind[j][0], r_ind[j][1]))).T[0] for j in range(nzones)]
                    sigma2_ind = [np.where((np.isfinite(error_ind[j])) & (error_unit*error_ind[j]<error_thres) & (error_ind[j]>0),
                                           (error_unit*error_ind[j])**2,
                                           0)
                                  for j in range(nzones)]
                    np_ind = [len(coord_list[i][ind[i]]) for ind in inds]
                    av_error.append([np.sqrt(np.nansum(sigma2_ind[j])/np_ind[j])/beams_zone_sqrt[i][j] for j in range(nzones) in np_ind])
                    
            else: #compute standard error of mean value 
                av_error = [np.array([np.std(resid_list[i][inds[j][i]], ddof=1)/beams_zone_sqrt[i][j] for i in range(nconts)]) for j in range(nzones)]

        return av_on_inds, av_error    

    
    def make_2d_map(self, prop_thres=np.inf, return_coords=False): 
        #compute x,y coords from azimuthal contours and interpolate onto 2D grid
        n_conts = len(self._coord_list)
        R_list = np.array([np.repeat(self._lev_list[i], len(self._coord_list[i]))
                           for i in range(n_conts)], dtype=object)
        R = np.concatenate(R_list).ravel()
        phi = np.radians(np.concatenate(self._coord_list).ravel())
        prop = np.concatenate(self._resid_list).ravel()
        prop[np.abs(prop)>prop_thres] = 0.0
        
        for p in [phi, R, prop]:
            p = np.nan_to_num(p, nan=0.0)
            
        x = R*np.cos(phi)
        y = R*np.sin(phi)
        prop2D = griddata((x, y), prop, (self.X, self.Y), method='linear')
        
        if return_coords:
            return x, y, prop, prop2D
        else:
            return prop2D

    def make_filaments(self, surface='upper', **kwargs):
        #FIND FILAMENTS
        #kwargs docs at https://fil-finder.readthedocs.io/en/latest/tutorial.html#masking

        from fil_finder import FilFinder2D

        kw_fil_mask = dict(
            verbose=False,
            border_masking=False,
            adapt_thresh=self.beam_size,
            smooth_size=0.2*self.beam_size,
            size_thresh=500*u.pix**2,
            fill_hole_size=0.01*u.arcsec**2
        )
        kw_fil_mask.update(kwargs)
        
        Rgrid = self.R_nonan[surface]        
        R_min_m=self.Rmin.to('m').value
        ang_scale = np.abs(self.header['CDELT1'])*u.Unit(self.header['CUNIT1']) #pix size
                       
        Rind = (Rgrid>R_min_m) #& (Rgrid<R_max)
        fil_pos = FilFinder2D(np.where(Rind & (self.prop>0), np.abs(self.prop), 0), ang_scale=ang_scale, distance=self.dpc)
        fil_pos.preprocess_image(skip_flatten=True) 
        fil_pos.create_mask(**kw_fil_mask)
        fil_pos.medskel(verbose=False)
        
        fil_neg = FilFinder2D(np.where(Rind & (self.prop<0), np.abs(self.prop), 0), ang_scale=ang_scale, distance=self.dpc)
        fil_neg.preprocess_image(skip_flatten=True) 
        fil_neg.create_mask(**kw_fil_mask)
        fil_neg.medskel(verbose=False)

        fil_pos.analyze_skeletons(prune_criteria='length')
        fil_neg.analyze_skeletons(prune_criteria='length')
        return fil_pos, fil_neg

    
class Contours(object):
    @staticmethod
    def emission_surface(ax, R, phi, extent, R_lev=None, phi_lev=None,
                         proj_offset=None, X=None, Y=None, which='both',
                         kwargs_R={}, kwargs_phi={}):
        kwargs_phif = dict(linestyles=':', linewidths=0.4, colors='k')
        kwargs_Rf = dict(linestyles=':', linewidths=0.4, colors='k')
        kwargs_phif.update(kwargs_phi)        
        kwargs_Rf.update(kwargs_R)

        near_nonan = ~np.isnan(R['upper'])

        Rmax = np.nanmax(R['upper'])

        kwargs_phif.update({'extent': extent})
        kwargs_Rf.update({'extent': extent})

        if R_lev is None: R_lev = np.linspace(0.06, 0.97, 4)*Rmax
        else: R_lev = np.sort(R_lev)
        if phi_lev is None: phi_lev = np.radians(np.arange(-170, 171, 20))

        #Splitting phi into pos and neg to try and avoid ugly contours close to -pi and pi
        phi_lev_neg = phi_lev[phi_lev<0] 
        phi_lev_pos = phi_lev[phi_lev>=0]
        phi_neg_near = np.where((phi['upper']<0) & (R['upper']>R_lev[0]) & (R['upper']<R_lev[-1]), phi['upper'], np.nan)
        phi_pos_near = np.where((phi['upper']>0) & (R['upper']>R_lev[0]) & (R['upper']<R_lev[-1]), phi['upper'], np.nan)
        phi_neg_far = np.where((phi['lower']<0) & (R['lower']>R_lev[0]) & (R['lower']<R_lev[-1]), phi['lower'], np.nan)
        phi_pos_far = np.where((phi['lower']>0) & (R['lower']>R_lev[0]) & (R['lower']<R_lev[-1]), phi['lower'], np.nan)

        if proj_offset is not None: #For 3d projections
            ax.contour(X, Y, R['upper'], offset=proj_offset, levels=R_lev, **kwargs_Rf)
            ax.contour(X, Y, np.where(near_nonan, np.nan, R['lower']), offset=proj_offset, levels=R_lev, **kwargs_Rf)
            ax.contour(X, Y, phi_pos_near, offset=proj_offset, levels=phi_lev_pos, **kwargs_phif)
            ax.contour(X, Y, phi_neg_near, offset=proj_offset, levels=phi_lev_neg, **kwargs_phif)
            ax.contour(X, Y, np.where(near_nonan, np.nan, phi_pos_far), offset=proj_offset, levels=phi_lev_pos, **kwargs_phif)
            ax.contour(X, Y, np.where(near_nonan, np.nan, phi_neg_far), offset=proj_offset, levels=phi_lev_neg, **kwargs_phif)
            
        else:
            if which=='both':
                ax.contour(R['upper'], levels=R_lev, **kwargs_Rf)
                ax.contour(np.where(R['upper']<R_lev[-1], np.nan, R['lower']), levels=R_lev, **kwargs_Rf)
                ax.contour(phi_pos_near, levels=phi_lev_pos, **kwargs_phif)
                ax.contour(phi_neg_near, levels=phi_lev_neg, **kwargs_phif)
                ax.contour(np.where(near_nonan, np.nan, phi_pos_far), levels=phi_lev_pos, **kwargs_phif)
                ax.contour(np.where(near_nonan, np.nan, phi_neg_far), levels=phi_lev_neg, **kwargs_phif)
            elif which in ['upper', 'up']:
                ax.contour(R['upper'], levels=R_lev, **kwargs_Rf)
                ax.contour(phi_pos_near, levels=phi_lev_pos, **kwargs_phif)
                ax.contour(phi_neg_near, levels=phi_lev_neg, **kwargs_phif)
            elif which in ['lower', 'low']:
                ax.contour(R['lower'], levels=R_lev, **kwargs_Rf)
                ax.contour(phi_pos_far, levels=phi_lev_pos, **kwargs_phif)
                ax.contour(phi_neg_far, levels=phi_lev_neg, **kwargs_phif)
                
    #The following method can be optimised if the contour finding process is separated from the plotting
    # by returning coords_list and inds_cont first, which will allow the user use the same set of contours to plot different props.

    @staticmethod
    def disc_axes(ax, R_list, z_list, incl, PA, xc=0, yc=0, **kwargs_axes):

        kwargs_ax = dict(color='k', ls=':', lw=1.5, dash_capstyle='round', dashes=(0.5, 1.5), alpha=0.7)
        kwargs_ax.update(kwargs_axes)
        
        phi_daxes_0 = np.zeros_like(R_list)
        phi_daxes_90 = np.zeros_like(R_list)+np.pi/2
        phi_daxes_180 = np.zeros_like(R_list)+np.pi
        phi_daxes_270 = np.zeros_like(R_list)-np.pi/2

        isort = np.argsort(R_list)
        make_ax = lambda x, y: ax.plot(x, y, **kwargs_ax)        
        
        for phi_dax in [phi_daxes_0, phi_daxes_90, phi_daxes_180, phi_daxes_270]:            
            x_cont, y_cont,_ = GridTools.get_sky_from_disc_coords(R_list[isort], phi_dax, z_list[isort], incl, PA, xc, yc)
            make_ax(x_cont, y_cont)

    @staticmethod
    def make_substructures(*args, **kwargs): #Backcompat
        __doc__ = make_substruct.__doc__
        return make_substruct(*args, **kwargs)
        
    @staticmethod
    def make_contour_lev(prop, lev, X, Y, acc_threshold=20): 
        contour = measure.find_contours(prop, lev)
        ind_good = np.argmin([np.abs(lev-prop[tuple(np.round(contour[i][0]).astype(int))]) for i in range(len(contour))]) #get contour id closest to lev              
        inds_cont = np.round(contour[ind_good]).astype(int)
        inds_cont = [tuple(f) for f in inds_cont]
        first_cont = np.array([prop[i] for i in inds_cont])
        corr_inds = np.abs(first_cont-lev) < acc_threshold
        x_cont = np.array([X[i] for i in inds_cont])
        y_cont = np.array([Y[i] for i in inds_cont])
        return x_cont[corr_inds], y_cont[corr_inds], inds_cont, corr_inds

    @staticmethod
    def beams_along_ring(lev, Rgrid, beam_size, X, Y):
        xc, yc, _, _ = Contours.make_contour_lev(Rgrid, lev, X, Y)
        try:
            rc = hypot_func(xc, yc)
            a = np.max(rc)
            b = np.min(rc)
            ellipse_perim = np.pi*(3*(a+b)-np.sqrt((3*a+b)*(a+3*b))) #Assumes that the disc vertical extent does not distort much the ellipse
            return ellipse_perim/beam_size
        except ValueError: #No contour found
            return np.inf

    @staticmethod
    def get_average_east_west(resid_list, coord_list, lev_list, 
                              Rgrid, beam_size, X, Y,
                              av_func=np.nanmean, mask_ang=0, resid_thres='3sigma',
                              error_func=True, error_unit=1.0, error_thres=np.inf):
        #resid_thres: None, '3sigma', or list of thresholds with size len(lev_list)        
        nconts = len(lev_list)
        if resid_thres is None: resid_thres = [np.inf]*nconts
        elif resid_thres == '3sigma': resid_thres = [3*np.nanstd(resid_list[i]) for i in range(nconts)] #anything higher than 3sigma is rejected from annulus
        # -np.pi<coord_list<np.pi
        ind_west = [((coord_list[i]<90-mask_ang) & (coord_list[i]>-90+mask_ang)) & (np.abs(resid_list[i]-np.nanmean(resid_list[i])) < resid_thres[i]) for i in range(nconts)]
        ind_east = [((coord_list[i]>90+mask_ang) | (coord_list[i]<-90-mask_ang)) & (np.abs(resid_list[i]-np.nanmean(resid_list[i])) < resid_thres[i]) for i in range(nconts)]
        av_west = np.array([av_func(resid_list[i][ind_west[i]]) for i in range(nconts)])
        av_east = np.array([av_func(resid_list[i][ind_east[i]]) for i in range(nconts)])
        
        if error_func is None: av_west_error, av_east_error = None, None
        else:
            beams_ring_sqrt = np.sqrt([0.5*Contours.beams_along_ring(lev, Rgrid, beam_size, X, Y) for lev in lev_list]) #0.5 because we split the disc in halves
            if callable(error_func): #if error map provided, compute average error per radius, divided by sqrt of number of beams (see Michiel Hogerheijde notes on errors)
                av_west_error, av_east_error = np.zeros(nconts), np.zeros(nconts)
                for i in range(nconts):
                    x_west, y_west, __ = get_sky_from_disc_coords(lev_list[i], coord_list[i][ind_west[i]]) #MISSING z, incl, PA for the function to work
                    x_east, y_east, __ = get_sky_from_disc_coords(lev_list[i], coord_list[i][ind_east[i]]) #MISSING z, incl, PA for the function to work
                    error_west = np.array(list(map(error_func, x_west, y_west))).T[0]
                    error_east = np.array(list(map(error_func, x_east, y_east))).T[0]
                    sigma2_west = np.where((np.isfinite(error_west)) & (error_unit*error_west<error_thres) & (error_west>0), (error_unit*error_west)**2, 0)
                    sigma2_east = np.where((np.isfinite(error_east)) & (error_unit*error_east<error_thres) & (error_east>0), (error_unit*error_east)**2, 0)
                    Np_west = len(coord_list[i][ind_west[i]])
                    Np_east = len(coord_list[i][ind_east[i]])
                    av_west_error[i] = np.sqrt(np.nansum(sigma2_west)/Np_west)/beams_ring_sqrt[i]  
                    av_east_error[i] = np.sqrt(np.nansum(sigma2_east)/Np_east)/beams_ring_sqrt[i]        
            else: #compute standard error of mean value
                av_west_error = np.array([np.std(resid_list[i][ind_west[i]], ddof=1) for i in range(nconts)])/beams_ring_sqrt
                av_east_error = np.array([np.std(resid_list[i][ind_east[i]], ddof=1) for i in range(nconts)])/beams_ring_sqrt
                
        return av_west, av_east, av_west_error, av_east_error

    @staticmethod
    def get_average(resid_list, coord_list, lev_list, 
                    Rgrid, beam_size, X, Y,
                    av_func=np.nanmean, mask_ang=0, resid_thres='3sigma',
                    error_func=True, error_unit=1.0, error_thres=np.inf):
        #mask_ang: +- angles to reject around minor axis (i.e. phi=+-90) 
        #resid_thres: None, '3sigma', or list of thresholds with size len(lev_list)        
        frac_annulus = 1.0 #if halves, 0.5; if quadrants, 0.25
        nconts = len(lev_list)
        if resid_thres is None: resid_thres = [np.inf]*nconts #consider all values for the average
        elif resid_thres == '3sigma': resid_thres = [3*np.nanstd(resid_list[i]) for i in range(nconts)] #anything higher than 3sigma is rejected from annulus
        # -np.pi<coord_list<np.pi        
        ind_accep = [(((coord_list[i]<90-mask_ang) & (coord_list[i]>-90+mask_ang)) |
                      ((coord_list[i]>90+mask_ang) | (coord_list[i]<-90-mask_ang))) &
                     (np.abs(resid_list[i]-np.nanmean(resid_list[i]))<resid_thres[i])
                     for i in range(nconts)]

        av_annulus = np.array([av_func(resid_list[i][ind_accep[i]]) for i in range(nconts)])
        
        if error_func is None: av_error = None
        else:
            beams_ring_sqrt = np.sqrt([frac_annulus*Contours.beams_along_ring(lev, Rgrid, beam_size, X, Y) for lev in lev_list])
            if callable(error_func): #if error map provided, compute average error per radius, divided by sqrt of number of beams (see Michiel Hogerheijde notes on errors)
                av_error = np.zeros(nconts)
                for i in range(nconts):
                    x_accep, y_accep, __ = get_sky_from_disc_coords(lev_list[i], coord_list[i][ind_accep[i]]) #MISSING z, incl, PA for the function to work
                    error_accep = np.array(list(map(error_func, x_accep, y_accep))).T[0]
                    sigma2_accep = np.where((np.isfinite(error_accep)) & (error_unit*error_accep<error_thres) & (error_accep>0), (error_unit*error_accep)**2, 0)
                    Np_accep = len(coord_list[i][ind_accep[i]])
                    av_error[i] = np.sqrt(np.nansum(sigma2_accep)/Np_accep)/beams_ring_sqrt[i]  
            else: #compute standard error of mean value
                av_error = np.array([np.std(resid_list[i][ind_accep[i]], ddof=1) for i in range(nconts)])/beams_ring_sqrt
                
        return av_annulus, av_error

    @staticmethod
    def get_average_zones(resid_list, coord_list, lev_list, Rgrid, beam_size, X, Y, fast=True, 
                          az_zones=[[-30, 30], [150,  -150]], join_zones=False, av_func=np.nanmean,
                          resid_thres='3sigma', error_func=True, error_unit=1.0, error_thres=np.inf):
                          
        #resid_thres: None, '3sigma', or list of thresholds with size len(lev_list)
        nconts = len(lev_list)
        nzones = len(az_zones)
        if not fast: join_zones=False
        
        if resid_thres is None: resid_thres = [np.inf]*nconts
        elif resid_thres == '3sigma': resid_thres = [3*np.nanstd(resid_list[i]) for i in range(nconts)] #anything higher than 3sigma is rejected from annulus

        make_or = lambda az0, az1: [((coord_list[i]>az0) | (coord_list[i]<az1)) & (np.abs(resid_list[i]-np.nanmean(resid_list[i])) < resid_thres[i]) for i in range(nconts)]
        make_and = lambda az0, az1: [((coord_list[i]>az0) & (coord_list[i]<az1)) & (np.abs(resid_list[i]-np.nanmean(resid_list[i])) < resid_thres[i]) for i in range(nconts)]

        def get_portion_inds(az):
            az0, az1 = az
            if (az0 > az1):
                inds = make_or(az0, az1)
            else:
                inds = make_and(az0, az1)
            return inds

        def get_portion_percent(az):
            az0, az1 = az
            if (az0 > az1):
                if az0 < 0: perc = 1 - (az0-az1)/360.
                else: perc = (180-az0 + 180-np.abs(az1))/360.
            else:
                perc = (az1-az0)/360.
            return perc

        #inds containts lists of indices, one list per zone. Each is a list of lists, with as many lists as nconts (number of radii). Each sublist has different number of indices, the larger the radius (i.e. larger path) the more indices.        
        inds = [get_portion_inds(zone) for zone in az_zones] 
        az_percent = np.array([get_portion_percent(zone) for zone in az_zones])

        if join_zones and nzones>1:
            concat = lambda x,y: x+y
            inds = [[functools.reduce(concat, [ind[i] for ind in inds]) for i in range(nconts)]] #concatenates indices from zones, per radius.
            az_percent = np.sum(az_percent)[None] #array of single number
            nzones = 1

        if fast: #Compute usual average
            av_on_inds = [np.array([av_func(resid_list[i][ind[i]]) for i in range(nconts)]) for ind in inds]        
        else: #Compute average using integral definition (trapezoid seems to succeed better than simpson)
            av_integral = lambda y,x,dT: np.trapz(y, x=x)/dT # or trapezoid from scipy.integrate
            av_on_inds = []
            for ind in inds:
                av_annulus = []
                for i in range(nconts):
                    ii = ind[i]
                    coords_ii = coord_list[i][ii]
                    resid_ii = resid_list[i][ii]
                    if not len(coords_ii): trap=None
                    else: trap = av_integral(resid_ii, coords_ii, coords_ii[-1]-coords_ii[0]) #dT assumes coords_list is sorted (no matter if ascending or descending)
                    av_annulus.append(trap) 
                av_on_inds.append(np.array(av_annulus))
            
        beams_ring_full = [Contours.beams_along_ring(lev, Rgrid, beam_size, X, Y) for lev in lev_list]
        beams_zone_sqrt = [np.sqrt(az_percent*br) for br in beams_ring_full]

        if error_func is None: av_error = None
        else:
            if callable(error_func): #Not yet tested
                #if error map provided, compute average error per radius, divided by sqrt of number of beams (see Michiel Hogerheijde notes on errors)  
                av_error = []
                for i in range(nconts):
                    r_ind = [get_sky_from_disc_coords(lev_list[i], coord_list[i][ind[i]]) for ind in inds] #MISSING z, incl, PA for the function to work
                    error_ind = [np.array(list(map(error_func, r_ind[j][0], r_ind[j][1]))).T[0] for j in range(nzones)]
                    sigma2_ind = [np.where((np.isfinite(error_ind[j])) & (error_unit*error_ind[j]<error_thres) & (error_ind[j]>0),
                                           (error_unit*error_ind[j])**2,
                                           0)
                                  for j in range(nzones)]
                    np_ind = [len(coord_list[i][ind[i]]) for ind in inds]
                    av_error.append([np.sqrt(np.nansum(sigma2_ind[j])/np_ind[j])/beams_zone_sqrt[i][j] for j in range(nzones) in np_ind])
            else: #compute standard error of mean value 
                av_error = [np.array([np.std(resid_list[i][inds[j][i]], ddof=1)/beams_zone_sqrt[i][j] for i in range(nconts)]) for j in range(nzones)]

        return av_on_inds, av_error    



"""
Useful repeating code
"""
import json
import decimal
import argparse
import warnings
import numbers
import runpy
import copy
import sys
import os

from discminer.plottools import get_discminer_cmap, make_1d_legend
from discminer.tools.utils import FrontendUtils
from discminer.grid import GridTools
from discminer.core import Data
from discminer.rail import Rail
from discminer.disc2d import Model
import discminer.cart as cart

from astropy.io import fits
import astropy.units as u

import numpy as np
import matplotlib


SMALL_SIZE = 10
MEDIUM_SIZE = 15

disctex = {
    'dmtau': r'$\rm DM\,Tau$',
    'aatau': r'$\rm AA\,Tau$',
    'lkca15': r'$\rm LkCa15$',
    'hd34282': r'$\rm HD\,34282$',
    'mwc758': r'$\rm MWC\,758$', 
    'cqtau': r'$\rm CQ\,Tau$',
    'sycha': r'$\rm SY\,Cha$',
    'pds66': r'$\rm PDS\,66$',
    'hd135344': r'$\rm HD\,135344B$',
    'j1604': r'$\rm J1604$',
    'j1615': r'$\rm J1615$',
    'v4046': r'$\rm V4046\,Sgr$',
    'j1842': r'$\rm J1842$',
    'j1852': r'$\rm J1852$',
    'hd143006': r'$\rm HD\,143006$',
    'hd163296': r'$\rm HD\,163296$'
}

moltex = {
    '12co': r'$^{\rm 12}$CO',
    '13co': r'$^{\rm 13}$CO',
    'cs': r'CS',
    'hcn': r'HCN',
    'c2h': r'C$_{\rm 2}$H'
}

warnings.filterwarnings("ignore", category=UserWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

def init_data_and_model(parfile='parfile.json', Rmin=0, Rmax=1.1, twodim=False, init_model=True, write_extent=True, verbose=True):
    #Rmin: If dimensionless, fraction of beam_size
    #Rmax: If dimensionless, fraction of Rout    
    with open(parfile) as jf:
        pars = json.load(jf)

    meta = pars['metadata']
    best = pars['best_fit']
    custom = pars['custom']

    file_data = meta['file_data']
    Rout = best['intensity']['Rout']
    dpc = meta['dpc']*u.pc

    datacube = Data(file_data, dpc, twodim=twodim, disc=meta['disc'], mol=meta['mol']) # Read data and convert to Cube object

    func_defaults = {
        'velocity_func': cart.keplerian_vertical,
        'z_upper_func': cart.z_upper_exp_tapered,
        'z_lower_func': cart.z_lower_exp_tapered,
        'intensity_func': cart.intensity_powerlaw_rout,
        'linewidth_func': cart.linewidth_powerlaw,
        'lineslope_func': cart.lineslope_powerlaw,
        'line_profile': cart.line_profile_bell,
        'line_uplow': cart.line_uplow_mask
    }
    
    def _check_customtags(model, func):

        if func == 'z_upper_func':            

            if 'surf2pwl' in meta['kind']:
                model.z_upper_func = cart.z_upper_powerlaw
            elif 'surfirregular' in meta['kind']:
                model.z_upper_func = cart.z_upper_irregular                
            else:
                model.z_upper_func = func_defaults[func]
            
        elif func == 'z_lower_func':            

            if 'surf2pwl' in meta['kind']:
                model.z_lower_func = cart.z_lower_powerlaw
            else:
                model.z_lower_func = func_defaults[func]

        elif func == 'intensity_func':
            
            if 'I2pwl' in meta['kind']:
                model.intensity_func = cart.intensity_powerlaw_rbreak
            elif 'I2pwlnosurf' in meta['kind']:
                model.intensity_func = cart.intensity_powerlaw_rbreak_nosurf    
            else:
                model.intensity_func = func_defaults[func]

        elif func == 'line_uplow':
            
            if 'sum' in meta['kind']:
                model.line_uplow = cart.line_uplow_sum
            else:
                model.line_uplow = func_defaults[func]
                
        else:
            setattr(model, func, func_defaults[func])
            
    def set_model_funcs(model):
        func_list = list(func_defaults.keys())        
        func_custom = dict(f for f in zip(func_list, [False]*len(func_list)))
            
        try: #Check if local customcart module exists
            cwd = os.getcwd()
            sys.path.append(cwd)
            import customcart
            from inspect import getmembers, isfunction
            funcs = getmembers(customcart, isfunction)

            for f in funcs:
                if f[0] in func_list:
                    setattr(model, f[0], f[1])
                    func_custom[f[0]] = True
                    
        except ImportError:
            print ('Using default and/or tag-based parametric forms for model...')

        for func in func_list:
            if not func_custom[func]:
                _check_customtags(model, func)

    if init_model:

        if isinstance(Rmax, numbers.Real):
            Rmax = Rmax*Rout*u.au
            
        model = Model(datacube, Rmax=Rmax, Rmin=Rmin, write_extent=write_extent, prototype=True, verbose=verbose)

        set_model_funcs(model)

        #****************
        #PROTOTYPE PARAMS
        #****************
        model.params = copy.copy(best)
        model.params['intensity']['I0'] /= meta['downsamp_factor']

        return datacube, model

    else:
        return datacube

    
def read_json(decimals=True): #Read json file again but keep track of (if) decimals
    if decimals:
        parse_func = lambda x: decimal.Decimal(str(x))
    else:
        parse_func = float

    with open('parfile.json') as json_file: 
        pars = json.load(json_file, parse_float=parse_func)
    return pars
            
def get_2d_plot_decorators(moment, unit_simple=False, fmt_vertical=False):    
    I_res2abs = 3.0
    v_res2abs = 20.0
    L_res2abs = 5.0
    Ls_res2abs = 3.0

    pars = read_json()
    vsys = float(pars['best_fit']['velocity']['vsys'])
    custom = pars['custom']
    meta = pars['metadata']

    try:
        ctitle = disctex[meta['disc']]+' '+moltex[meta['mol']]
        
    except KeyError:
        ctitle = meta['disc']+' '+meta['mol']

    if moment=='delta_velocity':
        clim = custom['vlim']
        fclim = float(clim)*1000
        unit = '[m s$^{-1}$]'
        if unit_simple: unit = 'm/s'        
        clabel = r'$\Delta$ Velocity %s'%unit
        cmap_mom = matplotlib.pyplot.get_cmap('magma')
        cmap_res = cmap_mom 
        levels_im = np.linspace(0.2*fclim, fclim, 64)
        levels_cc = np.linspace(0.2*fclim, fclim, 4)

    if moment=='delta_linewidth':
        clim = custom['Llim']
        fclim = float(clim)*1000
        unit = '[m s$^{-1}$]'
        if unit_simple: unit = 'm/s'        
        clabel = r'$\Delta$ Line width %s'%unit
        cmap_mom = 'magma' #get_discminer_cmap('linewidth')
        cmap_res = cmap_mom #get_discminer_cmap('linewidth', kind='residuals')
        levels_im = np.linspace(0.0, fclim, 64)
        levels_cc = np.linspace(0.2*fclim, fclim, 4)

    if moment=='delta_peakintensity':
        clim = custom['Ilim']
        fclim = float(clim)
        unit = '[K]'
        if unit_simple: unit = 'K'        
        clabel = r'$\Delta$ Peak Intensity %s'%unit
        cmap_mom = 'magma' #get_discminer_cmap('intensity_2')
        cmap_res = cmap_mom #get_discminer_cmap('intensity_2', kind='residuals')
        levels_im = np.linspace(0.0, fclim, 64)
        levels_cc = np.linspace(0.2*fclim, fclim, 4)
        
    if moment=='reducedchi2':
        clim = custom['Ilim']
        fclim = float(clim)
        unit = '[R-Chi2]'
        if unit_simple: unit = 'R-Chi2'        
        clabel = r'Reduced $X^2$'
        cmap_mom = 'magma'
        cmap_res = cmap_mom 
        levels_im = np.linspace(0.0, fclim, 64)
        levels_cc = np.linspace(0.2*fclim, fclim, 4)
        
    if moment in ['velocity', 'v0phi', 'v0r', 'v0z', 'vr_leftover']:
        clim = custom['vlim']
        fclim = float(clim)        
        unit = '[km s$^{-1}$]'
        if unit_simple: unit = 'km/s'
        clabel = r'Centroid Velocity %s'%unit
        cmap_mom = cmap_res = get_discminer_cmap('velocity')
        levels_im = np.linspace(-v_res2abs*fclim, v_res2abs*fclim, 64)+vsys
        levels_cc = np.linspace(-2, 2, 9)+vsys
    
    if moment=='linewidth':
        clim = custom['Llim']
        fclim = float(clim)
        unit = '[km s$^{-1}$]'
        if unit_simple: unit = 'km/s'        
        clabel = r'Line width %s'%unit
        cmap_mom = get_discminer_cmap('linewidth')
        cmap_res = get_discminer_cmap('linewidth', kind='residuals')
        levels_im = np.linspace(0.0, L_res2abs*fclim, 64)
        levels_cc = np.linspace(0.2, 0.8, 4)

    if moment=='lineslope':
        clim = custom['Lslim']
        fclim = float(clim)        
        unit = ''
        clabel = r'Line slope'
        cmap_mom = get_discminer_cmap('linewidth')
        cmap_res = get_discminer_cmap('linewidth', kind='residuals')
        levels_im = np.linspace(0.0, Ls_res2abs*fclim, 48)
        levels_cc = np.linspace(1.0, Ls_res2abs*fclim, 5)
    
    if moment in ['peakintensity', 'peakint']:
        clim = custom['Ilim']
        fclim = float(clim)        
        unit = '[K]'
        if unit_simple: unit = 'K'        
        clabel = r'Peak Intensity %s'%unit
        cmap_mom = get_discminer_cmap('intensity_2')    
        cmap_res = get_discminer_cmap('intensity_2', kind='residuals')
        levels_im = np.linspace(0.0, I_res2abs*fclim, 64)
        levels_cc = np.linspace(10, 50, 5)

    sp_dec = str(clim).split('.')
    nsign = 0
    if fmt_vertical:
        nsign = 1
    
    if len(sp_dec)==2:
        ndigs = len(sp_dec[0]) + len(sp_dec[-1])        
        cfmt = '%'+'%d.%df'%(ndigs+1+nsign, len(sp_dec[-1]))
    else:
        ndigs = len(sp_dec[0])
        cfmt = '%'+'%d.1f'%(ndigs+1+1+nsign) #ndigs + . + 1f + nsign

    if 'delta' in moment:
        cfmt = '%d'
        
    return ctitle, clabel, fclim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit

def get_1d_plot_decorators(moment, tag=''):
    I_res2abs = 5.0
    v_res2abs = 20.0
    L_res2abs = 7.0
    Ls_res2abs = 3.0
        
    pars = read_json(decimals=False)
    vsys = float(pars['best_fit']['velocity']['vsys'])
    custom = pars['custom']
    meta = pars['metadata']

    if len(tag)>0: tag = tag+' '
        
    if moment in ['velocity', 'v0phi', 'v0r', 'v0z', 'vr_leftover']:
        unit = '[km/s]' #'[km s$^{-1}$]'
        clim0_res = -custom['vlim']
        clim1_res = custom['vlim']    
        clim0, clim1 = 0.0, v_res2abs*clim1_res + vsys
        clabel = clabel_res = None #Modified within each script
        
    if moment=='linewidth':
        clim0_res = -custom['Llim']
        clim1_res = custom['Llim']
        clim0, clim1 = 0, L_res2abs*clim1_res
        unit = '[km/s]' #'[km s$^{-1}$]'
        clabel = r'Line width %s%s'%(tag, unit)
        clabel_res = r'L. width residuals %s'%unit    

    if moment=='lineslope':
        clim0_res = -custom['Lslim']
        clim1_res = custom['Lslim']
        clim0, clim1 = 0, Ls_res2abs*clim1_res
        unit = ''
        clabel = r'Line slope'
        clabel_res = r'L. slope residuals'
        
    if moment in ['peakintensity', 'peakint']:
        clim0_res = -custom['Ilim']
        clim1_res = custom['Ilim']
        clim0, clim1 = 0.0, I_res2abs*clim1_res
        unit = '[K]'
        clabel = r'Peak Intensity %s%s'%(tag, unit)
        clabel_res = r'Peak Int. residuals %s'%unit    
    
    return clabel, clabel_res, clim0, clim0_res, clim1, clim1_res, unit

def _get_mask_tuples(lims):
    mask_tup = []
    for i in range(len(lims[::2])):
        mask_tup.append(lims[i*2:i*2+2])
    return mask_tup

def _merge_R_phi_mask(mask_R, mask_phi):
    #Mask and merge azimuthal+radial sections

    mask_phi['lims_tup'] = _get_mask_tuples(mask_phi['lims'])
    mask_R['lims_tup'] = _get_mask_tuples(mask_R['lims'])

    mask_sec = np.zeros_like(mask_R['map2d']).astype(bool)

    if len(mask_phi['lims_tup'])==0:
        mask_phi['lims_tup'] = [None]*len(mask_R['lims_tup'])

    if len(mask_R['lims_tup'])==0:
        mask_R['lims_tup'] = [None]*len(mask_phi['lims_tup'])

    cross_sign = lambda x, y: x | y
    same_sign = lambda x, y: x & y
    
    for i in range(len(mask_phi['lims_tup'])):
        lim_phi = mask_phi['lims_tup'][i]
        lim_R = mask_R['lims_tup'][i]

        if (lim_phi is None) and (lim_R is None):
            continue

        elif (lim_phi is not None) and (lim_R is None):
            limdown, limup = lim_phi
            dcond = mask_phi['map2d'] >= limdown
            ucond = mask_phi['map2d'] <= limup            

            if limdown*limup < 0 and limdown > limup:
                mask_func = cross_sign
            else:
                mask_func = same_sign

            mask_sec = mask_sec | mask_func(dcond, ucond)

        elif (lim_phi is None) and (lim_R is not None):
            limdown, limup = lim_R
            mask_sec = mask_sec | ((mask_R['map2d'] >= limdown) & (mask_R['map2d'] <= limup))

        else: #Both are not None
            limdown_R, limup_R = lim_R
            limdown_phi, limup_phi = lim_phi
            dcond = mask_phi['map2d'] >= limdown_phi
            ucond = mask_phi['map2d'] <= limup_phi            

            if limdown_phi*limup_phi < 0 and limdown_phi > limup_phi:
                mask_func = cross_sign
            else:
                mask_func = same_sign
                
            mask_sec = mask_sec | (mask_func(dcond, ucond) & (mask_R['map2d'] >= limdown_R) & (mask_R['map2d'] <= limup_R))

    return mask_sec

def make_masks(ax, mask_R, mask_phi, Rmax=1000, **kwargs_mask):

    kw_mask = dict(edgecolor='none', facecolor='w', alpha=1)
    kw_mask.update(kwargs_mask)

    mask_phi_tup = _get_mask_tuples(mask_phi)
    if len(mask_phi_tup)==0:
        mask_phi_tup = [[0, 360]]

    mask_R_tup = _get_mask_tuples(mask_R)
    if len(mask_R_tup)==0:
        mask_R_tup = [[0, Rmax]]

    def add_mask(mask_phi_i, mask_R_i):
        # Adapted from https://stackoverflow.com/questions/22789356/plot-a-donut-with-fill-or-fill-between
        n, radii = 50, list(mask_R_i)
        theta = np.radians(np.linspace(*mask_phi_i, n, endpoint=True))
        xs = np.outer(radii, np.cos(theta))
        ys = np.outer(radii, np.sin(theta))

        # x,y coords should be traversed in opposite directions
        xs[0,:] = xs[0,::-1]
        ys[0,:] = ys[0,::-1]
        
        ax.fill(np.ravel(xs), np.ravel(ys), **kw_mask)
        ax.fill(np.ravel(xs), np.ravel(ys), edgecolor='k', facecolor='none', lw=1.5)    
        
    for i, maski in enumerate(mask_phi_tup):
        add_mask(mask_phi_tup[i], mask_R_tup[i])
            
def get_noise_mask(
        datacube,
        thres=4, return_mean=True,
        mask_phi={'map2d': None,
                  'lims': [] #List of limits
        },
        mask_R={'map2d': None,
                'lims': [] #List of limits
        }        
):    
    #std from line-free channels, per pixel
    data = datacube.data    
    noise = np.std( np.append(data[:5,:,:], data[-5:,:,:], axis=0), axis=0)

    if return_mean:
        noise = np.nanmean(noise)

    mask = np.nanmax(data, axis=0) < thres*noise
    mask_sec = _merge_R_phi_mask(mask_R, mask_phi)
    mask = mask | mask_sec
    
    return noise, mask


def load_moments(
        args, moment=None, kernel=None, mask=[],
        clip_Rgrid=None, clip_Rmin=0*u.au, clip_Rmax=np.inf*u.au,
        deltas=False
):    
    if moment is None:
        moment = args.moment

    if kernel is None:
        kernel = args.kernel
        
    if kernel in ['gauss', 'gaussian', 'quadratic', 'bell']:
        tag_surf = 'both'
        ref_surf = 'upper' #For contour plots
    else:
        if args.surface in ['up', 'upper']:
            tag_surf = 'up'
            ref_surf = 'upper'
        elif args.surface in ['low', 'lower']:
            tag_surf = 'low'
            ref_surf = 'lower'
            
    if kernel in ['gauss', 'gaussian']:
        tag_base = f'{moment}_gaussian'
    elif kernel in ['quadratic']:
        tag_base = f'{moment}_quadratic'        
    elif kernel in ['bell']:
        tag_base = f'{moment}_bell'
    elif kernel in ['dgauss', 'doublegaussian']:
        tag_base = f'{moment}_{tag_surf}_doublegaussian_{args.kind}'
    elif kernel in ['dbell', 'doublebell']:
        tag_base = f'{moment}_{tag_surf}_doublebell_{args.kind}'

    if deltas:
        tag_base = 'delta_'+tag_base
        
    #Read and mask moment maps, and compute residuals
    moment_data = fits.getdata(tag_base+'_data.fits').squeeze()
    moment_model = fits.getdata(tag_base+'_model.fits').squeeze()

    try:
        if args.smooth>0.0:
            from scipy.ndimage import gaussian_filter 
            moment_data = gaussian_filter(moment_data, args.smooth)
            moment_model = gaussian_filter(moment_model, args.smooth)
            
    except AttributeError:
        pass

    moment_data_unma = moment_data
    moment_model_unma = moment_model
    moment_data[mask] = np.nan
    moment_model[mask] = np.nan
    
    def clip_radially(prop2d): #should be within masking function
        Rmin = clip_Rmin.to('au').value
        Rgrid = np.nan_to_num(clip_Rgrid, nan=np.inf).to('au').value
        Rmax = clip_Rmax.to('au').value
        return np.where((Rgrid<Rmin) | (Rgrid>Rmax), np.nan, prop2d)    

    if isinstance(clip_Rgrid, u.Quantity):
        moment_data = clip_radially(moment_data)
        moment_model = clip_radially(moment_model)

    if moment in ['delta_velocity', 'delta_linewidth']:
        moment_data *= 1000
        moment_model *= 1000
        
    residuals = moment_data - moment_model

    if args.surface in ['low', 'lower']:
        pass
        #from scipy.ndimage import gaussian_filter
        #moment_data = gaussian_filter(moment_data, 0.5)
        #moment_data[moment_data > 50.0] = np.nan

    return moment_data, moment_model, residuals, dict(surf = tag_surf,
                                                      ref_surf = ref_surf,
                                                      base = tag_base,
                                                      mask = np.isnan(moment_data),
                                                      data_unmasked = moment_data_unma,
                                                      model_unmasked = moment_model_unma)


def load_disc_grid():
    R = dict(
        upper=np.load('upper_R.npy'),
        lower=np.load('lower_R.npy')
    )

    phi = dict(
        upper=np.load('upper_phi.npy'),
        lower=np.load('lower_phi.npy')
    )

    z = dict(
        upper=np.load('upper_z.npy'),
        lower=np.load('lower_z.npy')
    )
    return R, phi, z

def make_and_save_filaments(map2d,
                            model=None,
                            tag='',                            
                            writefits=True,
                            return_all=False,
                            cmap='jet',
                            surface='upper'
):

    if model is None:
        _, model = init_data_and_model()
        model.make_model()

    mrail = Rail(model, map2d)
    Rgrid = model.projected_coords['R'][surface]
    
    fil_pos, fil_neg = mrail.make_filaments(
        smooth_size=0.1*model.beam_size,
        adapt_thresh=0.5*model.beam_size,
        size_thresh=100*u.pix**2
    )

    """
    masked_pos = np.ma.masked_where(fil_pos.skeleton!=1, fil_pos.skeleton)
    masked_neg = np.ma.masked_where(fil_neg.skeleton!=1, fil_neg.skeleton)    
    ax.imshow(masked_pos, cmap='Reds_r', alpha=1, zorder=3, extent = [mm.X.min(),mm.X.max(), mm.Y.min(),mm.Y.max()], origin='lower')
    ax.imshow(masked_neg, cmap='Blues_r', alpha=1, zorder=3, extent = [mm.X.min(),mm.X.max(), mm.Y.min(),mm.Y.max()], origin='lower')
    """

    fil_pos_list = [fil_pos.skeleton, fil_pos.skeleton_longpath]
    fil_neg_list = [fil_neg.skeleton, fil_neg.skeleton_longpath]
    fil_pos_obj, fil_neg_obj = [], []

    check_outer_nans = lambda fil: np.isnan(Rgrid[fil.pixel_coords]).any()
        
    for i,fil in enumerate(fil_pos.filaments):
        
        if check_outer_nans(fil):
            continue
        else:
            fil_pos_obj.append(fil)
        
        fil_i = np.zeros_like(fil_pos.skeleton)
        fil_i[fil.pixel_coords] = 1 
        fil_pos_list.append(fil_i)

        
    for i,fil in enumerate(fil_neg.filaments):
        
        if check_outer_nans(fil):
            continue
        else:
            fil_neg_obj.append(fil)
        
        fil_i = np.zeros_like(fil_neg.skeleton)
        fil_i[fil.pixel_coords] = 1 
        fil_neg_list.append(fil_i)

    fil_pos.filaments = fil_pos_obj
    fil_neg.filaments = fil_neg_obj
    
    if writefits:
        if len(tag)>0:
            if tag[0]!='_':
                tag='_'+tag
                
        fits.writeto('filaments_pos%s.fits'%tag, np.asarray(fil_pos_list), header=model.header, overwrite=True)
        fits.writeto('filaments_neg%s.fits'%tag, np.asarray(fil_neg_list), overwrite=True)

    cmap = matplotlib.cm.get_cmap(cmap)
    npos = len(fil_pos.filaments)
    nneg = len(fil_neg.filaments)
    cpos = np.linspace(0.6, 1.0, npos)        
    cneg = np.linspace(0.4, 0.0, nneg)
    colors_dict = {}
    colors_dict.update({i+1: matplotlib.colors.to_hex(cmap(cpos[i])) for i in range(npos)})
    colors_dict.update({-i-1: matplotlib.colors.to_hex(cmap(cneg[i])) for i in range(nneg)})
        
    if return_all:
        return fil_pos, fil_neg, fil_pos_list, fil_neg_list, colors_dict       

    else:
        return fil_pos_list, fil_neg_list, colors_dict


def mark_planet_location(ax, args, r=[], phi=[], labels=[], coords='disc', zfunc=None, zpars=None, incl=None, PA=None, xc=None, yc=None, dpc=None, kwargs_text={}, **kwargs_scatter):

    kwargs_sc = dict(edgecolors='gold', facecolors='none', marker='o', s=450, lw=4.5, alpha=1.0, label=None, zorder=22)
    kwargs_sc.update(kwargs_scatter)

    kwargs_tx = dict(fontsize=MEDIUM_SIZE-2, color='k', ha='center', va='center', weight='bold', rotation=0, zorder=23)
    kwargs_tx.update(kwargs_text)

    if len(r)==0:
        rs = args.rp
        phis = args.phip
    else:
        rs = r
        phis = phi

    npoints = len(rs)    

    if len(labels)==0:
        labels = args.labelp

    if len(labels)==1:
        labels = labels*npoints

    xcoords, ycoords = [], []
    for i in range(npoints):

        if args.input_coords in ['disc', 'disk']:
            if coords=='disc': #Coordinate system for the planet marker
                if args.projection=='cartesian':
                    phii = np.radians(phis[i])
                    xi, yi = rs[i]*np.cos(phii), rs[i]*np.sin(phii)
            
                elif args.projection=='polar':
                    xi, yi = phis[i], rs[i]

            elif coords=='sky':
                phii = np.radians(phis[i])
                zp = zfunc({'R': rs[i]*u.au.to('m')}, **zpars)*u.m.to('au')
                xi,yi,zi = GridTools.get_sky_from_disc_coords(rs[i], phii, zp, incl, PA, xc, yc) 

        elif args.input_coords=='sky':
            rsky = rs[i]*(dpc.to('pc')).value #Projected distance in au
            phisky = np.radians(phis[i]) #Position angle, measured from the North
            xsky = -rsky*np.sin(phisky)
            ysky = rsky*np.cos(phisky)
            
            if coords=='disc':
                xdisc, ydisc = GridTools.get_disc_from_sky_coords(xsky, ysky, zfunc, zpars, incl, PA, xc=xc, yc=yc, midplane=True)
                
                if args.projection=='cartesian':
                    xi, yi = xdisc, ydisc
                    ri = np.hypot(xdisc, ydisc)
                    print (ri)
                    
                elif args.projection=='polar':
                    phii = np.arctan2(ydisc, xdisc)
                    ri = np.hypot(xdisc, ydisc)
                    xi, yi = np.degrees(phii), ri

            elif coords=='sky':
                xi, yi = xsky, ysky
                
        ax.scatter(xi, yi, **kwargs_sc)
        
        if len(labels)>0:
            ax.text(xi, yi, labels[i], **kwargs_tx)

        xcoords.append(xi)
        ycoords.append(yi)
        
    return xcoords, ycoords

def overlay_filaments(ax, x, y, prop, projection='cartesian', **kwargs):

    fil_pos_obj, fil_neg_obj, *_ = make_and_save_filaments(prop, return_all=True, **kwargs)

    if projection=='cartesian':
        ax.contour(x, y, fil_pos_obj.skeleton, linewidths=0.2, colors='darkred', alpha=1, zorder=11)
        ax.contour(x, y, fil_neg_obj.skeleton, linewidths=0.2, colors='navy', alpha=1, zorder=11)

    if projection=='polar':
        proj_fil = lambda fil_obj: (x[fil_obj.skeleton.astype(bool)], y[fil_obj.skeleton.astype(bool)])
        ax.scatter(*proj_fil(fil_pos_obj), marker='o', lw=0.9, s=20, ec='k', fc='tomato', alpha=1, zorder=11)
        ax.scatter(*proj_fil(fil_neg_obj), marker='o', lw=0.9, s=20, ec='k', fc='dodgerblue', alpha=1, zorder=11)

    
    
def overlay_continuum(ax, parfile='parfile_scattered.json', comp=-1, convert_to_tb=False, vmin=0.1, vmax=1.0, lev=32, contours=False, coords='disc', surface='upper', data_cont=None, model_cont=None, model_proj=None, **kwargs_contourf):

    if data_cont is None or model_cont is None:
        data_cont, model_cont = init_data_and_model(parfile=parfile, twodim=True, write_extent=False)

    if convert_to_tb:
        try:
            data_cont.convert_to_tb(writefits=False)
        except KeyError:
            warnings.warn('\nUnable to convert continuum intensity to brightness temperature units...\n', Warning)
    
    if model_proj is None:
        R_cont, phi_cont, z_cont, R_nonan, phi_nonan, z_nonan = model_cont.get_projected_coords(writebinaries=False)
    else:
        R_cont, phi_cont, z_cont, R_nonan, phi_nonan, z_nonan = model_proj
    

    img = data_cont.data[comp]
    peak_img = np.nanmax(img)
    kwargs = dict(levels=np.linspace(vmin*peak_img, vmax*peak_img, lev))
    kwargs.update(kwargs_contourf)

    Xproj_cont = R_cont[surface]*np.cos(phi_cont[surface])*u.m.to('au')
    Yproj_cont = R_cont[surface]*np.sin(phi_cont[surface])*u.m.to('au')

    if contours:
        levels=np.linspace(vmin*peak_img, vmax*peak_img, lev)                        
        kwargs.update(levels=levels)
        linewidths = 2.0*levels/np.max(levels)

        if coords=='sky':        
            im = ax.contour(img, extent=model_cont.extent, levels=levels, colors='k', linewidths=linewidths, zorder=30) 
        if coords in ['disc', 'disk']:
            im = ax.contour(Xproj_cont, Yproj_cont, img, levels=levels, colors='k', linewidths=linewidths, zorder=30) 

    if coords=='sky':
        im = ax.contourf(img, extent=model_cont.extent, **kwargs)
    if coords in ['disc', 'disk']:
        im = ax.contourf(Xproj_cont, Yproj_cont, img, **kwargs)
            
    return im, data_cont, model_cont


def overlay_spirals(ax, args, mtags, Rmin=0, Rmax=np.inf, phi_extent=2*np.linspace(-360, 360, 500)):

    #SPIRAL PRESCRIPTIONS
    sp_lin = lambda x, a, b: a + b*x
    sp_log = lambda x, a, b: a*np.exp(b*x)
    
    arr_read = np.loadtxt(
        'spirals_fit_parameters_%s_%s.txt'%(mtags['base'].replace(args.moment, args.spiral_moment), args.spiral_type),
        dtype = {
            'names': ('a', 'b', 'color', 'sign', 'id'),
            'formats': (float, float, 'U10', 'U10', int)
        },
        skiprows = 1,
        comments = None,
    )

    arr_read = np.atleast_1d(arr_read)
    phi_ext_rad = np.radians(phi_extent)

    if args.spiral_type == 'linear':
        sp_func = sp_lin
    else:
        sp_func = sp_log
            
    for arr in arr_read:
        if (
                (arr[3]=='pos' and arr[4] in args.spiral_ids)
                or
                (arr[3]=='neg' and arr[4] in args.spiral_ids)
        ):
            #Can this be translated back to sky coordinates?
            R_ext = sp_func(phi_ext_rad, *tuple(arr)[:2])
            ind = (R_ext > Rmin) & (R_ext < Rmax)
            xplot = R_ext[ind]*np.cos(phi_ext_rad[ind])
            yplot = R_ext[ind]*np.sin(phi_ext_rad[ind])            
            ax.plot(xplot, yplot, lw=3.5,  color=arr[2], zorder=20)
            ax.plot(xplot, yplot, lw=4.5, color='k', zorder=19)

def save_fits(data, header, overwrite=True, momenttype='velocity', filename='map.fits'):

    hdr = copy.copy(header)
    if momenttype=='velocity':
        hdr["BUNIT"] = "km/s"
        hdr["BTYPE"] = "Velocity"

    kwargs = dict(overwrite=overwrite, header=hdr)

    fits.writeto(filename, data, **kwargs)

def show_output(args):
    if args.show_output:
        matplotlib.pyplot.show(block=args.show_block)
    else:
        matplotlib.pyplot.close()



from discminer.mining_control import _mining_channels
from discminer.mining_utils import get_noise_mask, init_data_and_model, show_output

from discminer.cube import Cube
from discminer.plottools import use_discminer_style

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import json
import copy
import sys

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_channels(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************    
with open('parfile.json') as json_file:
    pars = json.load(json_file)
    
meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

chan_step = custom['chan_step']
nchans = custom['nchans']

file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model(Rmin=0, Rmax=1.2)
vchannels = datacube.vchannels
pix_downsamp = model.grid['step']*meta['downsamp_fit']/au_to_m

#Useful definitions for plots
xmax = model.skygrid['xmax'] 
xlim = 1.0*xmax/au_to_m
extent= np.array([-xmax, xmax, -xmax, xmax])/au_to_m
  
#**************************
#MAKE MODEL (2D ATTRIBUTES)
#**************************
#Return model cube and compute disc coordinates projected on the sky
if args.make_beam==-1:
    if meta['mol'] in ['12co', '13co'] and pix_downsamp>1.01*model.beam_size.value:
        modelcube = model.make_model(make_convolve=False) 
    else:
        modelcube = model.make_model(make_convolve=True)
else:
    modelcube = model.make_model(make_convolve=True*args.make_beam)     

modelcube.filename = 'cube_model_%s.fits'%tag
modelcube.writefits() #Jy/bm
modelcube.convert_to_tb(writefits=True, planck=args.planck) #K

datacube.filename = 'cube_data_%s.fits'%tag
datacube.writefits()
datacube.convert_to_tb(writefits=True, planck=args.planck)

#**********************
#VISUALISE CHANNEL MAPS
#**********************
if args.show_output:
    modelcube.show(compare_cubes=[datacube], extent=extent, int_unit='Intensity [K]', show_beam=True, surface_from=model)
    modelcube.show_side_by_side(datacube, extent=extent, int_unit='Intensity [K]', show_beam=True,  surface_from=model)

#*****************
#PLOT CHANNEL MAPS
#*****************    
idlim = int(0.5*chan_step*(nchans-1))
plot_channels = np.linspace(-idlim,idlim,nchans) + np.argmin(np.abs(vchannels-best['velocity']['vsys']))  #Channel ids to be plotted, selected around ~vsys channel

#DATA CHANNELS
fig, ax, im, cbar = datacube.make_channel_maps(channels={'indices': plot_channels}, ncols=5, contours_from=model)
plt.savefig('channel_maps_data.png', bbox_inches = 'tight', dpi=200)
plt.close()

#MODEL CHANNELS
fig, ax, im, cbar = modelcube.make_channel_maps(channels={'indices': plot_channels}, ncols=5, levels=im[0].levels, contours_from=model)
plt.savefig('channel_maps_model.png', bbox_inches = 'tight', dpi=200)
plt.close()

#RESIDUAL CHANNELS
noise_mean, mask = get_noise_mask(datacube)

residualscube = Cube(datacube.data-modelcube.data, datacube.header, datacube.vchannels, dpc, beam=datacube.beam)
residualscube.filename = 'cube_residuals_%s.fits'%tag
residualscube.writefits() 

fig, ax, im, cbar = residualscube.make_channel_maps(channels={'indices': plot_channels}, ncols=5,
                                                    kind='residuals',
                                                    contours_from=model,
                                                    projection=None, 
                                                    unit_intensity='Kelvin',
                                                    unit_coordinates='au',
                                                    xlims = (-xlim, xlim),
                                                    ylims = (-xlim, xlim),
                                                    mask_under=3*noise_mean,
                                                    levels=np.linspace(-custom['Ilim'], custom['Ilim'], 32))


#SHOW DISC EMISSION SURFACE AND MAIN AXES
ic, jc = (np.asarray(np.shape(ax))/2).astype(int)
model.make_emission_surface(
    ax[ic][jc],
    kwargs_R={'colors': '0.4', 'linewidths': 0.4},
    kwargs_phi={'colors': '0.4', 'linewidths': 0.3}
)

model.make_disc_axes(ax[ic][jc])

"""
for axi in ax[0]:
    model.make_emission_surface(
        axi,
        kwargs_R={'colors': '0.4', 'linewidths': 0.4},
        kwargs_phi={'colors': '0.4', 'linewidths': 0.3}
    )
    
for axi in ax[1]:
    model.make_disc_axes(axi)
"""    

plt.savefig('channel_maps_residuals.png', bbox_inches = 'tight', dpi=200)
plt.close()




from .units import amu
#***************
#constants in SI
#***************
mp = 1.00727647*amu #kg ->proton mass 
mH = 1.00794*amu #kg ->hydrogen mass
G = 6.67384e-11 #(N m^2 kg^-2)Cavendish-gravitational constant
kb = 1.38064852e-23 #(J K^-1) Boltzmann constant
sigma = 5.670373e-8 #(W m^-2 K^-4)Stefan-Boltzmann constant
temp_cmb = 2.72548 #(K) CMB black-body temperature
c = 299792458. #(m/s) Speed of light
h = 6.62607015e-34 #(J Hz^-1) Planck constant



from discminer.core import Data
import astropy.units as u

file_data = 'MWC_480_CO_220GHz.robust_0.5.JvMcorr.image.pbcor'
dpc = 162.0*u.pc

#**********************
#DATACUBE FOR PROTOTYPE
#**********************
datacube = Data(file_data+'.fits', dpc)

datacube.clip(npix=250,  overwrite=True) # can also clip along the velocity axis using e.g. channels={"interval": [15, 115]})
datacube.downsample(2, tag='_2pix') # Downsample cube and add tag to filename

#**********************
#DATACUBE FOR MCMC FIT
#**********************
datacube = Data(file_data+'_clipped.fits', dpc)
datacube.downsample(10, tag='_10pix')



from discminer.mining_control import _mining_residuals_all
from discminer.mining_utils import (init_data_and_model,
                                    get_2d_plot_decorators,
                                    get_noise_mask,
                                    load_moments,
                                    load_disc_grid,
                                    mark_planet_location,
                                    show_output)

from discminer.rail import Contours
from discminer.plottools import (add_cbar_ax,
                                 make_substructures,
                                 _make_radial_grid_2D,
                                 make_up_ax,
                                 mod_minor_ticks,
                                 mod_major_ticks,
                                 use_discminer_style)

import json
import numpy as np
from astropy import units as u
import matplotlib.pyplot as plt

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_residuals_all(None)
    args = parser.parse_args()

if args.coords=='disk':
    args.coords = 'disc'
    
#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*args.Router*Rout*u.au

#*******************
#LOAD DATA AND GRID
#*******************
#datacube = Data(file_data, dpc) # Read data and convert to Cube object
datacube, model = init_data_and_model(Rmin=0, Rmax=Rmax)
noise_mean, mask = get_noise_mask(datacube)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.05*np.min([xmax, Rmax.value]) 
extent= np.array([-xmax, xmax, -xmax, xmax])

R, phi, z = load_disc_grid()

R_nonan_au = np.nan_to_num(R[args.surface])/au_to_m
z_nonan_au = np.nan_to_num(z[args.surface])/au_to_m

Xproj = R_nonan_au*np.cos(phi[args.surface])
Yproj = R_nonan_au*np.sin(phi[args.surface])

nh = int(0.5*datacube.nx)

#***********************************
#LOAD MOMENT MAPS AND GET RESIDUALS
#***********************************
moments_data = {}
moments_model = {}
residuals = {}
mtags = {}
mtypes = ['linewidth', 'velocity', 'peakintensity']

for moment in mtypes:
    if args.coords=='sky':
        md, mm, rr, mt = load_moments(args, moment=moment, mask=mask)
    elif args.coords=='disc':
        md, mm, rr, mt = load_moments(
            args,
            moment=moment,
            mask=mask,
            clip_Rmin=args.Rinner*datacube.beam_size,
            clip_Rmax=args.Router*Rout*u.au,
            clip_Rgrid=R[args.surface]*u.m
        )        
    mtags[moment] = mt
    moments_data[moment] = md 
    moments_model[moment] = mm 
    residuals[moment] = rr
         
#*****************
#HELPER FUNCTIONS
#*****************
def decorate_ax_res_2D(ax, cbar, lim=xlim):
    ax.set_aspect(1)
    make_up_ax(ax, xlims=(-lim,lim), ylims=(-lim,lim), labelsize=13)
    ax.set_xlabel('Offset [au]')
    ax.xaxis.set_label_position('top')    
    cbar.ax.tick_params(which='major', direction='in', width=2.7, size=4.8, pad=4, labelsize=12)
    cbar.ax.tick_params(which='minor', direction='in', width=2.3, size=3.3)
    mod_minor_ticks(cbar.ax)

#***************
#PLOT RESIDUALS
#***************    
fig, ax = plt.subplots(ncols=3, nrows=1, figsize=(17,6))
cbar_axes = [add_cbar_ax(fig, axi, perc=5, pad=0) for axi in ax]

clabels = {
    'linewidth': r'$\Delta$ Linewidth [km s$^{-1}$]',
    'velocity': r'$\Delta$ Centroid Velocity [km s$^{-1}$]',
    'peakintensity': r'$\Delta$ Peak Intensity [K]'
}
fill_angs_2pi = np.linspace(0, 2*np.pi, 100)

for i, (axi, moment) in enumerate(zip(ax, mtypes)):
    ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(moment)
    levels_resid = np.linspace(-clim, clim, 32)
    levels_cbar = np.linspace(-clim, clim, 5)

    chot = '0.5'
    cmap_mom.set_under(chot)
    cmap_mom.set_over(chot)
    
    cmap_res.set_under(chot)
    cmap_res.set_over(chot)
    
    kwargs_im = dict(cmap=cmap_res, levels=levels_resid, extend='both', origin='lower')
    kwargs_cbar = dict(orientation='horizontal', format=cfmt, ticks=levels_cbar, pad=0.03, shrink=0.95, aspect=15)    
    
    if args.coords=='sky':
        im = axi.contourf(residuals[moment], extent=extent, **kwargs_im)
        Contours.emission_surface(axi, R, phi, extent=extent,
                                  R_lev=np.linspace(0.1, 1.0, 10)*Rout*au_to_m,
                                  which=mtags['velocity']['surf']
        )
        Contours.disc_axes(axi,
                           R[args.surface][nh]/au_to_m,
                           z[args.surface][nh]/au_to_m,
                           incl, PA, xc=xc, yc=yc)

        Contours.emission_surface(
            axi, R, phi, extent=extent,
            R_lev=np.array(rings)*u.au.to('m'), which='upper',
            kwargs_R={'linestyles': '-', 'linewidths': 1.2, 'colors': 'k'},
            kwargs_phi={'colors': 'none'}
        )
        
    elif args.coords=='disc':
        im = axi.contourf(Xproj, Yproj, residuals[moment], **kwargs_im)
        #make_substructures(axi, gaps=gaps, rings=rings, twodim=True)
        if Rout>700: lf = 4
        else: lf = 2
        _make_radial_grid_2D(axi, args.Router*Rout, gaps=gaps, make_labels=True, label_freq=lf)
        
    cbar = plt.colorbar(im, cax=cbar_axes[i], **kwargs_cbar)
    cbar.set_label(clabels[moment], fontsize=14)

    decorate_ax_res_2D(axi, cbar)
    datacube.plot_beam(axi, fc='lime')

ax[0].set_ylabel('Offset [au]')

tag_base = mtags['velocity']['base'].split('velocity_')[-1]

#*************
#MARK PLANETS
#*************
kwargs_sc = dict(s=150, lw=2.0, edgecolors='k')
for axi in ax:
    mark_planet_location(axi, args, dpc=dpc, coords='sky', zfunc=model.z_upper_func, zpars=best['height_upper'], **best['orientation'], **kwargs_sc)

plt.savefig('residuals_all_%s_%sframe.png'%(tag_base, args.coords), bbox_inches='tight', dpi=200)
show_output(args)



from discminer.mining_control import _mining_channels_peakint
from discminer.mining_utils import (
    get_noise_mask,
    get_2d_plot_decorators,
    init_data_and_model,
    load_moments,
    show_output
)

from discminer.cube import Cube
from discminer.plottools import (
    use_discminer_style,
    make_up_ax,
    get_continuous_cmap
)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import json

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_channels_peakint(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
Rout = best['intensity']['Rout']

nchans = args.nchans
chan_step = args.step

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

mol_tex = {
    '12co': r'$^{12}$CO',
    '13co': r'$^{13}$CO',
    'cs': r'CS'
}

try:
    mol_tex_meta = mol_tex[meta['mol']]
except KeyError:
    mol_tex_meta = meta['mol']

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model(Rmin=0, Rmax=1.0)
noise_mean, mask = get_noise_mask(datacube, thres=args.sigma)
    
vchannels = datacube.vchannels
pix_downsamp = model.grid['step']*meta['downsamp_fit']/au_to_m

#Useful definitions for plots
xmax = model.skygrid['xmax'] 
"""
if meta['mol'] in ['13co', 'cs']:
    xlim = 0.8*xmax/au_to_m
else:
    xlim = xmax/au_to_m
"""
xlim = args.Router*Rout    
extent= np.array([-xmax, xmax, -xmax, xmax])/au_to_m

#**************************
#MAKE MODEL (2D ATTRIBUTES)
#**************************
modelcube = model.make_model(make_convolve=True)
modelcube.convert_to_tb(writefits=False)
datacube.convert_to_tb(writefits=False)

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

#*****************
#PLOT CHANNEL MAPS
#*****************
noise_mean, mask = get_noise_mask(datacube, thres=args.sigma) #again, but in Tb units
print ('Mean RMS: %.2e K'%noise_mean)

cmap = None
"""
if 'sum' in meta['kind']:
    import cmasher as cmr
    cmap = cmap_mom = plt.get_cmap('cmr.rainforest_r')
    cmap = plt.get_cmap('cmr.rainforest_r')
"""
idlim = int(0.5*chan_step*(nchans-1))
plot_channels = np.linspace(-idlim,idlim,nchans) + np.argmin(np.abs(vchannels-best['velocity']['vsys'])) 
plot_channels = np.append([0], plot_channels)

kw_channels = dict(channels={'indices': plot_channels}, ncols=nchans+1,
                   xlims = (-xlim, xlim), ylims = (-xlim, xlim), show_beam=False,
                   observable = 'intensity_2',
                   #contours_from=model,
                   #kwargs_contour = dict(colors='w', linewidths=1.0), 
                   unit_coordinates='au', unit_intensity=None, projection=None)
                   
fig, ax = plt.subplots(nrows=3, ncols=nchans+1, figsize=(2.5*(nchans+1), 2.5*3)) #Master figure
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.05, hspace=0.05)

#DATA CHANNELS
levels=np.linspace(0*args.sigma*noise_mean, 3.5*custom['Ilim'], 32)

fig, ax0, im0, cbar0 = datacube.make_channel_maps(fig=fig, ax=ax[0,:], levels=levels, extend='max', **kw_channels)
fig, ax1, im1, cbar1 = modelcube.make_channel_maps(fig=fig, ax=ax[1,:], levels=im0[0].levels, annotate_channels=False, cmap=cmap, extend='max', **kw_channels)
cbar0.remove()

#RESIDUAL CHANNELS
kw_channels.update(dict(kwargs_contour = {}, contours_from=model))

residualscube = Cube(datacube.data-modelcube.data, datacube.header, datacube.vchannels, dpc, beam=datacube.beam)

fig, ax2, im2, cbar2 = residualscube.make_channel_maps(fig=fig, ax=ax[2,:],
                                                       kind='residuals',
                                                       annotate_channels=False,
                                                       extend = 'both',
                                                       mask_under=args.sigma*noise_mean,
                                                       levels=np.linspace(-custom['Ilim'], custom['Ilim'], 32),
                                                       #levels=12*np.linspace(-noise_mean, noise_mean, 32),
                                                       **kw_channels)


for axi in ax[:-1,0]:
    axi.set_xlabel('')    
    axi.tick_params(labelleft=False, labelbottom=False)

ax[0,0].set_title('Peak Intensity', color='0.6', fontsize=13, pad=8, loc='center')
ax[0,0].set_ylabel(mol_tex_meta + ' Data', color='k', fontsize=13, labelpad=6)
ax[1,0].set_ylabel('Model', color='k', fontsize=13, labelpad=4)

#PLOT MOMENT MAP
for axi in ax[:,0]:
    axi.set_xlim(-xlim, xlim)
    axi.set_ylim(-xlim, xlim)
    for artist in axi.get_lines() + axi.collections + axi.texts:
        artist.remove()

datacube.plot_beam(ax[0,0], fc='gold')        
        
kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=im0[0].levels[::1])

pm0 = ax[0,0].contourf(moment_data, extend='both', **kwargs_im)
pm1 = ax[1,0].contourf(moment_model, extend='both', **kwargs_im)
pm2 = ax[2,0].contourf(residuals, cmap=cmap_res, origin='lower', extend='both', extent=extent, levels=np.linspace(-clim, clim, 32))

#for axi in ax[:,int(round(0.5*(nchans+1)))-1]:
for axi in ax[:,0]:
    model.make_emission_surface(
        axi,
        kwargs_R={'colors': '0.1', 'linewidths': 0.5},
        kwargs_phi={'colors': '0.1', 'linewidths': 0.4}
    )
    model.make_disc_axes(axi)
    for spine in ['left',  'top', 'right', 'bottom']:
        axi.spines[spine].set_color('0.8')        
    axi.tick_params(which='both', color='0.8')
    #axi.axis('off')

plt.savefig('channel_maps_peakint_%s_%s.png'%(meta['disc'], meta['mol']), bbox_inches = 'tight', dpi=200)    
show_output(args)



def get_model_parameters():
    params = {
        'velocity': {
            'Mstar': 1.0,
            'vel_sign': 1,
            'vsys': 0
        },
        'orientation': {
            'incl': 0.7,
            'PA': 0.0,
            'xc': 0.0,
            'yc': 0.0
        },
        'intensity': {
            'I0': 0.5,
            'p': -1.5, 
            'q': 2.0,
            'Rout': 500
        },
        'linewidth': {
            'L0': 0.3, 
            'p': -0.5, 
            'q': -0.3
        }, 
        'lineslope': {
            'Ls': 2.0, 
            'p': 0.3, 
            'q': 0.0
        },
        'height_upper': {
            'z0': 40.0,
            'p': 1.0,
            'Rb': 500,
            'q': 2.0
        },
        'height_lower': {
            'z0': 20.0,
            'p': 1.0,
            'Rb': 500,
            'q': 2.0
        }
    }

def get_model_functions(disc=None, depth='thick', mol='12co'):

    model_funcs = copy.copy(_func_defaults)
    
    if disc in [None, 'mwc480']:
        z_upper_func = z_upper_exp_tapered
        z_lower_func = z_lower_exp_tapered
        intensity_func = intensity_powerlaw_rout
        linewidth_func = linewidth_powerlaw
        lineslope_func = lineslope_powerlaw
        
    if depth=='thick':
        model_funcs.update({'line_profile': line_profile_bell})
        line_uplow = line_uplow_mask

    elif depth=='thin':
        line_profile = line_profile_gaussian
        line_uplow = line_uplow_sum
    
#On INIT MODEL
#Could be the other way around, initialise datacube=Cube(...), inherit Model class which receives datacube and other parameters. Then all required info for metadata is in self object. If I decide to stick to the current version perhaps I should rename this as ReferenceCube or ModelCube instead?



from discminer.core import Data
import discminer.cart as cart
from discminer.plottools import (make_round_map,
                                 make_polar_map,
                                 make_substructures,
                                 make_up_ax,
                                 mod_minor_ticks,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)

from utils import (make_and_save_filaments,
                   make_1d_legend,
                   init_data_and_model,
                   get_2d_plot_decorators,
                   get_noise_mask,
                   load_moments,
                   load_disc_grid,
                   add_parser_args)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u
from scipy.optimize import curve_fit
from scipy.signal import savgol_filter

import json
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot residuals+filaments', description='Plot residual maps and overlay identified filaments')
parser.add_argument('-sppos', '--spirals_pos', nargs='*', default=[], type=int, help="Positive spiral ids to overlay fitted curve and save fit parameters into txt file. FORMAT: -sp 0 2 3")
parser.add_argument('-spneg', '--spirals_neg', nargs='*', default=[], type=int, help="Negative spiral ids to overlay fitted curve and save fit parameters into txt file. FORMAT: -sn 0 2 3")
parser.add_argument('-sptype', '--spiral_type', default='linear', choices=['linear', 'log'], help="Type of spiral fit to be shown and saved into file.")

args = add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, projection=True, Rinner=1, Router=True, smooth=True)

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

#********************
#LOAD DATA AND GRID
#********************
datacube = Data(file_data, dpc) # Read data and convert to Cube object
noise_mean, mask = get_noise_mask(datacube, thres=2)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()
RR = R[args.surface]/au_to_m
PP = phi[args.surface]

Xproj = RR*np.cos(phi[args.surface])
Yproj = RR*np.sin(phi[args.surface])

#*************************
#LOAD AND CLIP MOMENT MAPS    
moment_data, moment_model, residuals, mtags = load_moments(
    args,
    mask=mask,
    clip_Rmin=0.0*u.au,
    clip_Rmax=args.Router*Rout*u.au,
    clip_Rgrid=R[args.surface]*u.m
)

#******************************
#INIT MODEL AND MAKE FILAMENTS
#******************************
_, model = init_data_and_model()
model.make_model()
fil_pos_obj, fil_neg_obj, fil_pos_list, fil_neg_list, colors_dict = make_and_save_filaments(model, residuals, tag=mtags['base']+'_'+args.projection, return_all=True)
fil_pos = fil_pos_list[0]
fil_neg = fil_neg_list[0]

#***********
#MAKE PLOTS
clabels = {
    'linewidth': r'$\Delta$ Line width [km s$^{-1}$]',
    'lineslope': r'$\Delta$ Line slope',
    'velocity': r'$\Delta$ Centroid [km s$^{-1}$]',
    'peakintensity': r'$\Delta$ Peak Int. [K]'
}


fig, ax = plt.subplots(ncols=2, nrows=2, figsize=(14, 7))

axr = fig.add_axes([0.55, 0.1, 0.4, 0.8])
ax[0,1].set_visible(False)
ax[1,1].set_visible(False)

axf = ax[0,0] #fit in polar coords
axp = ax[1,0] #pitch angle

#FIT SPIRALS
sp_lin = lambda x, a, b: a + b*x
sp_log = lambda x, a, b: a*np.exp(b*x)

def make_savgol(prof):
    tmp = len(prof)
    wl = tmp if tmp%2 else tmp-1

    try:
        ysav = savgol_filter(prof, wl, 2)
        ysav_deriv = savgol_filter(prof, wl, 2, deriv=1)
    except np.linalg.LinAlgError:
        ysav = prof
        ysav_deriv = None

    return ysav, ysav_deriv

def sort_xy(x, xp, yp):
    #Sort xp, yp based on input x 
    ind = np.argsort(x)
    return xp[ind], yp[ind]

def clean_filament_xy(xp, yp):
    xp, yp = sort_xy(xp, xp, yp)
    dx = np.abs(xp[1:] - xp[:-1])
    ind = np.append([True], [dx > 0.5*np.median(dx)])
    xp = xp[ind]
    yp = yp[ind]
    
    def shift(xp, delta, ind):
        indp = np.arange(1, len(xp))[ind] #Indices with high delta
        i = indp[np.argmax(delta[indp-1])] #Take index with highest delta
        
        left = xp[:i]
        right = xp[i:]

        if len(left)<=len(right):
            xp[:i] += 360
        else:
            xp[i:] -= 360
            
        return xp

    #Attempt to connect substructure split by phi-axis discontinuity
    dy = np.abs(yp[1:] - yp[:-1])    
    #Consider 80% smallest deltas for statistics
    iy = np.argsort(dy)[:-int(0.2*len(dy))] 
    indj = dy > np.nanmean(dy[iy]) + 10*np.nanstd(dy[iy]) 
    if np.sum(indj):
        xp = shift(xp, dy, indj)        
        xp, yp = sort_xy(xp, xp, yp)

    #Check dx too
    dx = np.abs(xp[1:] - xp[:-1])
    #Consider 80% smallest deltas for statistics        
    ix = np.argsort(dx)[:-int(0.2*len(dx))] 
    indc = dx > np.nanmean(dx[ix]) + 10*np.nanstd(dx[ix])     
    if np.sum(indc):
        xp = shift(xp, dx, indc)
        xp, yp = sort_xy(xp, xp, yp)            

    #Remove 10% outliers
    #iy = np.argsort(yp)[:-np.ceil(0.1*len(yp)).astype(int)]
    #xp, yp = sort_xy(xp[iy], xp[iy], yp[iy])

    return xp, yp


def fit_and_plot(ax, xp, yp, color=None, kind='positive'):
    ind, = np.where(np.isnan(xp) | np.isnan(yp) | (yp < args.Rinner*datacube.beam_size.value))
    xp = np.delete(xp, ind)
    xp_rad = np.radians(xp)
    yp = np.delete(yp, ind)    
        
    if kind in ['pos', 'positive']:
        if color is None:
            color = 'r'
        lin_c = 'magenta'
        log_c = 'k'
        sav_c = 'tomato'
    elif kind in ['neg', 'negative']:
        if color is None:
            color = 'b'
        lin_c = 'cyan'
        log_c = 'k'
        sav_c = 'dodgerblue'
    else:
        raise ValueError
   
    ax.scatter(xp, yp, s=20, color=color)
    x_ext = 20*np.linspace(-np.pi, np.pi, 500)
    
    #LINEAR FIT
    popt_lin, pcov_lin = curve_fit(sp_lin, xp_rad, yp)
    ylin = sp_lin(xp_rad, *popt_lin)    
    ylin_ext = sp_lin(x_ext, *popt_lin) #Linear extrapolation

    #LOGARITHMIC FIT
    popt_log, pcov_log = curve_fit(sp_log, xp_rad, yp, p0=[100, 0])
    ylog = sp_log(xp_rad, *popt_log)    
    ylog_ext = sp_log(x_ext, *popt_log) #Log extrapolation        

    #print (np.sqrt(np.diag(pcov)))
    
    #ysav, dy = make_savgol(yp)
    #ax.plot(xp, ysav, c=sav_c, ls='-', lw=2)

    #DERIVATIVES
    dylin = np.gradient(ylin, xp_rad)
    dylog = np.gradient(ylog, xp_rad)

    dylin_ext = np.gradient(ylin_ext, x_ext)
    dylog_ext = np.gradient(ylog_ext, x_ext)
    
    #*********
    #MAKE BINS    
    #Estimate optimal nbins
    # using Rice rule or Freedman-Diaconis rule nbins=2*iqr*n**(-1/3)
    #  NOT BEING USED ATM
    r"""
    iqr = np.abs(np.quantile(xp, 0.75) - np.quantile(xp, 0.25)) 
    nbins = int(2*len(xp)**(1/3.)) #int(2*iqr*len(xp)**(-1/3.))

    from scipy import stats
    bin_stats = stats.binned_statistic
    xm, _, _ = bin_stats(xp, xp, statistic='mean', bins=nbins)
    ym, _, _ = bin_stats(xp, yp, statistic='mean', bins=nbins)
    ystd, _, _ = bin_stats(xp, yp, statistic='std', bins=nbins)        

    ind = ~np.isnan(xm)
    xm = xm[ind]
    ym = ym[ind]
    ystd = ystd[ind]

    #popt, pcov = curve_fit(sp_log, xm, ym, p0=[100, 0], sigma=ystd)     
    #ax.scatter(xm, ym, ec='k', fc='none', marker='o', lw=2, s=30)        
    #"""    
    #*********    

    if args.spiral_type == 'linear':
        ax.scatter(xp, ylin, ec='k', fc='none', marker='s', lw=0.3, s=15)        
        return ylin, dylin, ylin_ext, dylin_ext, popt_lin, popt_log
    else:
        ax.scatter(xp, ylog, ec='k', fc='none', marker='s', lw=0.3, s=15)        
        return ylog, dylog, ylog_ext, dylog_ext, popt_lin, popt_log

#***********
#LEFT PANELS
#***********
#POSITIVE SPIRALS
ipos, ineg = [], [] #Filament ids considered in the analysis
popt_lin_pos, popt_log_pos = [], []
popt_lin_neg, popt_log_neg = [], []

for i,fil in enumerate(fil_pos_list[2:]):    
    fil = fil #*fil_pos_list[1]
    fp = fil.astype(bool)
    xp = np.degrees(PP[fp])
    yp = RR[fp]
    xp, yp = clean_filament_xy(xp, yp)
    yfp, dyp, yp_ext, dyp_ext, popt_lin, popt_log = fit_and_plot(axf, xp, yp, color=colors_dict[i+1])
    pitchp_ext = np.degrees(np.arctan(np.abs(dyp_ext)/yp_ext))
    pitchp = np.degrees(np.arctan(np.abs(dyp)/yfp))

    popt_lin_pos.append(popt_lin)
    popt_log_pos.append(popt_log)
    
    if np.sum(fp)>=10:
        axp.plot(yp_ext, pitchp_ext, lw=2.5, ls='--', c=colors_dict[i+1])    
        axp.plot(yfp, pitchp, lw=4.0, c=colors_dict[i+1])
        ipos.append(i)
                
#NEGATIVE SPIRALS        
for i,fil in enumerate(fil_neg_list[2:]):    
    fil = fil #*fil_neg_list[1]
    fn = fil.astype(bool)
    xn = np.degrees(PP[fn])
    yn = RR[fn]
    xn, yn = clean_filament_xy(xn, yn)
    yfn, dyn, yn_ext, dyn_ext, popt_lin, popt_log = fit_and_plot(axf, xn, yn, color=colors_dict[-i-1], kind='neg')
    pitchn_ext = np.degrees(np.arctan(np.abs(dyn_ext)/yn_ext))    
    pitchn = np.degrees(np.arctan(np.abs(dyn)/yfn))

    popt_lin_neg.append(popt_lin)
    popt_log_neg.append(popt_log)
    
    if np.sum(fn)>=10:
        axp.plot(yn_ext, pitchn_ext, lw=2.5, ls='--', c=colors_dict[-i-1])    
        axp.plot(yfn, pitchn, lw=4.0, c=colors_dict[-i-1])
        ineg.append(i)
        
axp.plot([None], [None], lw=4.0, ls='-', c='k', label='Measurement')        
axp.plot([None], [None], lw=2.5, ls='--', c='k', label='Extrapolation')
make_1d_legend(axp, fontsize=11)

mod_major_ticks(axp, axis='y', nbins=5)
axp.set_xlim(args.Rinner*datacube.beam_size.value, args.Router*Rout)
#axp.set_ylim(-9, 61)
axp.set_ylim(-3, 43)
axp.set_xlabel('Radius [au]', labelpad=5)
axp.set_ylabel('Pitch angle [deg]')
make_substructures(axp, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)  
mod_major_ticks(axp, axis='x', nbins=10)
mod_major_ticks(axp, axis='y', nbins=5)

axf.set_ylim(0, None)
axf.set_ylim(args.Rinner*datacube.beam_size.value, args.Router*Rout)
axf.set_xlabel('Azimuth [deg]', labelpad=5)
axf.set_ylabel('Radius [au]')
axf.set_xticks(np.arange(-270,270+1,90))
for ang in [-90, 90, -270, 270]:
    axf.axvline(ang, ls=':', lw=2.5, color='0.3', dash_capstyle='round')
#axf.set_xlim(-1*360, 1*360)
axf.set_xlim(-1*181, 1*181)
    
axf.plot([None], [None], ls=':', lw=2.5, color='0.3', dash_capstyle='round', label='Minor axis')
make_1d_legend(axf)

#***********
#RIGHT PANEL
#***********
if args.projection=='cartesian':
    levels_resid = np.linspace(-clim, clim, 32)
    
    if args.surface in ['up', 'upper']:
        z_func = cart.z_upper_exp_tapered
        z_pars = best['height_upper']

    elif args.surface in ['low', 'lower']:
        z_func = cart.z_lower_exp_tapered
        z_pars = best['height_lower']
    
    fig, axr = make_round_map(residuals, levels_resid, Xproj*u.au, Yproj*u.au, args.Router*Rout*u.au,
                              fig=fig, ax=axr, make_cbar=False,
                              z_func=z_func, z_pars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc,
                              cmap=cmap_res, clabel=unit, fmt=cfmt, 
                              gaps=gaps, rings=rings,
                              mask_inner=args.Rinner*datacube.beam_size,
                              kwargs_mask={'zorder': 30})
                              #kwargs_mask={'facecolor': '0.6', 'alpha': 0.4})
    
    make_substructures(axr, gaps=gaps, rings=rings, twodim=True, label_rings=True)

    for i,fil in enumerate(fil_pos_list[2:]):
        axr.contour(Xproj, Yproj, fil, linewidths=1.0, colors=colors_dict[i+1], alpha=1, zorder=13)
    for i,fil in enumerate(fil_neg_list[2:]):        
        axr.contour(Xproj, Yproj, fil, linewidths=1.0, colors=colors_dict[-i-1], alpha=1, zorder=13)

    phi_ext = 4*np.linspace(-np.pi, np.pi, 500)

    if args.spiral_type == 'linear':
        sp_func = sp_lin
        popt_pos = popt_lin_pos
        popt_neg = popt_lin_neg
        header_txt = 'a\tb\tcolor\tsign\tid --> Linear fit: R = a + b*phi'
    else:
        sp_func = sp_log
        popt_pos = popt_log_pos
        popt_neg = popt_log_neg        
        header_txt = 'a\tb\tcolor\tsign\tid --> Logarithmic fit: R = a*np.exp(b*phi)'
        
    arr_write = []    
    for i in args.spirals_pos:
        R_ext = sp_func(phi_ext, *popt_pos[i])
        ind = (R_ext > 0) & (R_ext < args.Router*Rout)
        xplot = R_ext[ind]*np.cos(phi_ext[ind])
        yplot = R_ext[ind]*np.sin(phi_ext[ind])
        axr.plot(xplot, yplot, lw=2,  color=colors_dict[i+1], zorder=20)
        axr.plot(xplot, yplot, lw=5, color='k', zorder=19)

        arr_write.append(np.append(np.round(popt_pos[i], 6), [colors_dict[i+1], 'pos', i]))
        
    for i in args.spirals_neg:
        R_ext = sp_func(phi_ext, *popt_neg[i])
        ind = (R_ext > 0) & (R_ext < args.Router*Rout)
        xplot = R_ext[ind]*np.cos(phi_ext[ind])
        yplot = R_ext[ind]*np.sin(phi_ext[ind])        
        axr.plot(xplot, yplot, lw=2,  color=colors_dict[-i-1], zorder=20)
        axr.plot(xplot, yplot, lw=5, color='k', zorder=19)

        arr_write.append(np.append(np.round(popt_neg[i], 6), [colors_dict[-i-1], 'neg', i]))

    if len(arr_write)>0:
        arr_write = np.asarray(arr_write)
        np.savetxt(
            'spirals_fit_parameters_%s_%s.txt'%(mtags['base'], args.spiral_type),
            arr_write,
            fmt='%s',
            header=header_txt,
            delimiter='\t'
        )

        
axr.set_title(ctitle, fontsize=16, color='k')

plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.5)

plt.savefig('residuals_filaments_%s_%s_%s.png'%(mtags['base'], args.projection, args.spiral_type), bbox_inches='tight', dpi=200)
plt.show()
plt.close()

#**************
#WIDTH ANALYSIS
#**************
fil_pos_obj.find_widths()
fil_neg_obj.find_widths()
    
ncols = np.max([len(ipos), len(ineg)]).astype(int)

fig, ax = plt.subplots(nrows=2, ncols=ncols, figsize=(4*ncols, 8))
ax[0] = np.atleast_1d(ax[0])
ax[1] = np.atleast_1d(ax[1])


def filament_radprof(ax_cols, filaments, colors=None):
    data = []    
    for i in range(ncols):
        axi = ax_cols[i]            
        if i >= len(filaments):
            axi.set_visible(False)
            continue
        fili = filaments[i]
        try:
            fili.plot_radial_profile(xunit=u.au, ax=axi)
        except:
            print('Radial profile fit failed for Filament #%d'%i)
            axi.set_visible(False)

        lines = axi.get_lines()            

        width, error = fili.radprof_fwhm(u.au)
        for line in lines:
            if line.get_color()=='k' and colors is not None:
                line.set_markerfacecolor(colors[i])
                line.set_markeredgecolor('k')
                line.set_markersize(7)                
                data.append(np.max(line.get_data()[1]))

            else:
                line.set_linewidth(4.5)
                line.set_color('0.5')
                line.set_alpha(0.5)
                line.set_label('FWHM: %.1f +/-  %.1f au'%(width.value, error.value))
        axi.grid(ls='--')
        make_up_ax(axi, labelbottom=True, labeltop=False, labelsize=11)
        make_1d_legend(axi, fontsize=12)
        
    data_max = np.max(data)
    for i in range(ncols): 
        axi = ax_cols[i]
        axi.set_ylim(-np.round(0.05*data_max, 2), np.round(1.2*data_max, 2))
        

filament_radprof(ax[0], np.asarray(fil_pos_obj.filaments)[ipos], colors=[colors_dict[i+1] for i in ipos])
filament_radprof(ax[1], np.asarray(fil_neg_obj.filaments)[ineg], colors=[colors_dict[-i-1] for i in ineg])

ax00 = ax[0][0]
x0, x1 = ax00.get_xlim()
y0, y1 = ax00.get_ylim()
xp = 0.5*(x0+x1)
yp = 0.7*(y0+y1)
dx = datacube.beam_size.value #/(x1-x0)
ax00.text(0.5 + 0.5*dx/(x1-x0), 0.75, 'beam (%d au)'%datacube.beam_size.value, ha='center', fontsize=12, transform=ax00.transAxes)
ax00.plot([xp, xp+dx], [yp, yp], ls='-', lw=4.5, color='k')


if ncols==1:
    for axi in ax:
        axi[0].set_ylabel(clabels[args.moment], fontsize=12)

else:
    for axi in ax[:,0]:
        axi.set_ylabel(clabels[args.moment], fontsize=12)

    for i in range(2):
        for axi in ax[i,1:]:
            axi.set_ylabel(None)        
            axi.tick_params(labelleft=False)
            
for axi in ax[0]:
    axi.set_xlabel(None)    
for axi in ax[1]:
    axi.set_xlabel('Distance from spine [au]', fontsize=12)

plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.2, hspace=0.25)

plt.savefig('residuals_filaments_widths_%s_%s_%s.png'%(mtags['base'], args.projection, args.spiral_type), bbox_inches='tight', dpi=200)
plt.show()
plt.close()



from discminer.plottools import (make_round_map,
                                 make_polar_map,
                                 make_substructures,
                                 make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)

from utils import (load_disc_grid,
                   init_data_and_model,
                   get_noise_mask,
                   add_parser_args,
                   get_2d_plot_decorators,
                   load_moments)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u
import json

from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot gradient residual maps', description='Plot gradient residual maps')
args = add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, Rinner=True, Router=True, gradient=True, smooth=True)

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

levels_resid = np.linspace(-clim, clim, 48)

clim_grad = 0.04*clim #For gradient: dv/dR [unit/au]
cmap_grad = plt.get_cmap('nipy_spectral')
cmap_grad_r = plt.get_cmap('nipy_spectral_r')

levels_grad = np.linspace(0.05*clim_grad, clim_grad, 48)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube, thres=2)
vchannels = datacube.vchannels
model.make_model()

R, phi, z = load_disc_grid()

#*************************
#LOAD MOMENT MAPS    
Rin_plt = args.Rinner*datacube.beam_size
Rout_plt = args.Router*Rout*u.au

moment_data, moment_model, residuals, mtags = load_moments(
    args,
    mask=mask,
    clip_Rmin=0.0*u.au,
    clip_Rmax=Rout*u.au,
    clip_Rgrid=R[args.surface]*u.m
)

#*******************************
#DECORATIVE STUFF

if args.moment in ['velocity', 'linewidth']:
    unit = 'm/s'
    ufac = u.Unit('km/s').to(unit)
    dfmt = '%3d'
elif args.moment=='peakintensity':
    unit = 'K'
    ufac = 1
    dfmt = '%4.1f'
else:
    unit = ''
    ufac = 1
    dfmt = '%4.1f'
    
clabels = {
    'linewidth': r'$\Delta$ Line width [m s$^{-1}$]',
    'lineslope': r'$\Delta$ Line slope',
    'velocity': r'$\Delta$ Centroid [m s$^{-1}$]',
    'peakintensity': r'$\Delta$ Peak Int. [K]'
}

clabel_ = clabels[args.moment].split(' [')
clabel_unit = unit #clabel_[1].split(']')[0]
clabel_prop = clabel_[0]#.split('$')[-1]
cu = (clabel_prop, clabel_unit)

clabels_grad = {
    'r': r'$\delta/\delta R$ (%s) [%s/au]'%cu,
    'phi': r'$\delta/R\delta \phi$ (%s) [%s/au]'%cu,
    'peak': r'$\delta/\delta_{\rm peak}$ (%s) [%s/au]'%cu,
}

#**********
#MAKE PLOTS
#**********

if args.gradient=='peak':
    fig, ax = plt.subplots(ncols=1, nrows=2, figsize=(12, 6.5))
else:
    fig, ax = plt.subplots(ncols=1, nrows=3, figsize=(12, 10))

#Positive gradient    
_, _, cbar = make_polar_map(ufac*residuals, ufac*levels_grad,
                            R[args.surface]*u.m, phi[args.surface]*u.rad, Rout_plt,
                            fig=fig, ax=ax[0],
                            Rin=Rin_plt, gradient=args.gradient,
                            findpeaks='pos', filepeaks='phi_gradient_peaks_%s_positive.txt'%mtags['base'],
                            cmap=cmap_grad, fmt=dfmt, clabel=clabels_grad[args.gradient])
if args.gradient!='peak':
    #Negative gradient
    _, _, cbar = make_polar_map(ufac*residuals, ufac*np.sort(-levels_grad),
                                R[args.surface]*u.m, phi[args.surface]*u.rad, Rout_plt,
                                fig=fig, ax=ax[1],
                                Rin=Rin_plt, gradient=args.gradient,
                                findpeaks='neg', filepeaks='phi_gradient_peaks_%s_negative.txt'%mtags['base'],
                                cmap=cmap_grad_r, fmt=dfmt, clabel=clabels_grad[args.gradient])


#REFERENCE MAP
_, _, cbar = make_polar_map(ufac*residuals, ufac*levels_resid,
                            R[args.surface]*u.m, phi[args.surface]*u.rad, Rout_plt,
                            fig=fig, ax=ax[-1],
                            Rin=Rin_plt, gradient=0,
                            cmap=cmap_res, fmt='%4d', clabel=clabels[args.moment])


if args.moment=='velocity':
    try:
        peaks_pos = np.loadtxt('phi_gradient_peaks_%s_positive.txt'%mtags['base']).T    
        peaks_neg = np.loadtxt('phi_gradient_peaks_%s_negative.txt'%mtags['base']).T
        ax[-1].scatter(peaks_pos[0], peaks_pos[1], marker='+', color='indigo', lw=4, s=250)
        ax[-1].scatter(peaks_neg[0], peaks_neg[1], marker='_', color='indigo', lw=4, s=250)
        
    except FileNotFoundError:
        pass

for axi in ax:
    if axi!=ax[-1]:
        axi.set_xlabel(None)
        kc = {'color': '1.0'}
    else:
        kc = {}

    if axi==ax[-1]:
        label_rings = True
    else:
        axi.set_ylabel(None)
        label_rings = False
    make_substructures(axi, gaps=gaps, rings=rings, twodim=True, polar=True, label_rings=label_rings,
                       kwargs_gaps=kc, kwargs_rings=kc)

ax[0].set_title(ctitle, fontsize=16, color='k')

plt.savefig('gradient_residuals_deproj_%s_d%s_polar.png'%(mtags['base'], args.gradient), bbox_inches='tight', dpi=200)
plt.show()
plt.close()



from .cube import Cube
from .grid import grid as dgrid 
from .tools.utils import InputError

import numpy as np

from astropy.io import fits
from astropy import units as u
from radio_beam import Beam, Beams
from radio_beam.beam import NoBeamException
from spectral_cube import SpectralCube, VaryingResolutionSpectralCube

import numbers
import warnings
import json

class Data(Cube):
    def __init__(self, filename, dpc, twodim=False, disc=None, mol='12co', kind=['mask']):
        """
        Initialise Data object. Inherits `~discminer.disc2d.Cube` properties and methods.
        
        Parameters
        ----------
        filename : path-like 
            Path to input FITS file with datacube information.
        
        dpc : `~astropy.units.Quantity`
            Distance to the disc.
        """

        if twodim: #continuum image

            hdu = fits.open(filename)[0]
            header = hdu.header            
            data = hdu.data

            if len(data.shape)>3:
                data = data.squeeze()
            if len(data.shape)==2:
                data = np.expand_dims(data, 0)
                
            try:
                beam = Beam.from_fits_header(header)
            except NoBeamException:
                beam = None
                warnings.warn('No beam was found in the input FITS file.')
            try:
                ncomp = header['NAXIS3'] #e.g. Stokes components
            except KeyError:
                ncomp = 2

            if ncomp<2:
                ncomp = 2
                
            vchannels = np.arange(ncomp)

        else: #line cube
            cube_spe = SpectralCube.read(filename)
            try:
                cube_vel = cube_spe.with_spectral_unit(
                    u.km / u.s,
                    velocity_convention="radio",
                    rest_value=cube_spe.header["RESTFRQ"] * u.Hz,
                )
            except KeyError:
                #Assume velocity axis is already in km/s
                cube_vel = cube_spe
            
            #In km/s, remove .value to keep astropy units
            vchannels = cube_vel.spectral_axis.value
            header = cube_vel.header
        
            #Get data and beam info
            if isinstance(cube_vel, SpectralCube):
                data = cube_vel.hdu.data.squeeze()
                try:
                    beam = Beam.from_fits_header(header)  # radio_beam object
                except NoBeamException:
                    beam = None
                    warnings.warn('No beam was found in the input FITS file.')
            
            elif isinstance(cube_vel, VaryingResolutionSpectralCube):
                data = cube_vel.hdulist[0].data.squeeze()
                """
                beams = cube_vel.hdulist[1].data #One beam per channel
                bmaj = np.median(beams['BMAJ'])
                bmin = np.median(beams['BMIN'])
                bpa = np.median(beams['BPA'])            
                beam = Beam(major=bmaj*u.arcsec, minor=bmin*u.arcsec, pa=bpa*u.deg)
                """
                beams = Beams.from_fits_bintable(cube_vel.hdulist[1])
                beam = beams.common_beam() #Smallest common beam
                header.update(beam.to_header_keywords()) #Add single beam to header
            else:
                raise InputError(cube_vel,
                                 'The input datacube is not valid. Only the following spectral_cube instances are supported: SpectralCube, VaryingResolutionSpectralCube.')
            
        super().__init__(data, header, vchannels, dpc, beam=beam, filename=filename, disc=disc, mol=mol, kind=kind)

    def _init_cube(self):
        super().__init__(
            self.data,
            self.header,
            self.vchannels,
            self.dpc,
            beam = self.beam,
            filename = self.filename            
        )

"""
beam : None or `~radio_beam.Beam`, optional
- If None, the beam information is extracted from the header of the input FITS file.
- If `~radio_beam.Beam` object, it uses this 
"""
        
class ModelGrid():
    def __init__(self, datacube, Rmax, Rmin=1.0, write_extent=True):
        """
        Initialise ModelGrid object.

        Parameters
        ----------
        datacube : `~discminer.disc2d.cube.Cube` object
            Datacube to get sky grid from.

        Rmax : `~astropy.units.Quantity`
            Maximum radial extent of the model in physical units. Not to be confused with the disc outer radius.

        Rmin : float or `~astropy.units.Quantity`
            Inner radius to mask out from the model.

            - If float, computes inner radius in number of beams. Default is 1.0.

            - If `~astropy.units.Quantity`, takes the value provided, assumed in physical units.
        
        write_extent : bool
            If True, writes information about grid physical extent into JSON file.

        Attributes
        ----------
        skygrid : dict
            Dictionary with useful information of the sky grid where the disc observable properties are merged for visualisation. This grid matches the spatial grid of the input datacube in physical units.

        grid : dict
            Disc grid where the model disc properties are computed. Why are there two different grids? Sometimes, the maximum (deprojected) extent of the disc is larger than the rectangular size of the sky grid. Therefore, having independent grids is needed to make sure that the deprojected disc properties do not display sharp boundaries at R=skygrid_extent. In other cases, the maximum extent of the disc is smaller than the sky grid, in which case it's useful to employ a (smaller) independent grid to save computing time.
            
        """
        
        self.dpc = datacube.dpc
        self.Rmax = Rmax
        self.write_extent = write_extent
        
        if isinstance(datacube, Cube):
            self.datacube = datacube
            self.header = datacube.header
            self.vchannels = datacube.vchannels
            self.beam = datacube.beam            
            self._make_grid()
        else:
            raise InputError(datacube,
                             'The input cube must be a ~discminer.disc2d.cube.Cube instance.')
        
        if isinstance(Rmin, numbers.Real):
            if datacube.beam is not None:
                beam_au = (self.dpc*np.tan(datacube.beam.major.to(u.radian))).to(u.au)
                self.Rmin = Rmin*beam_au
            else: self.Rmin = 0.0*u.Unit(u.au)
        else:
            self.Rmin = Rmin.to(u.au)
            
        #elif isinstance(datacube, grid):
        #   self.beam = beam
        #   make_header()
        #   make_channels()
        #   make_prototype()
        #
        
        #if prototype:
        #   make_prototype()
        #       vel2d, int2d, linew2d, lineb2d = model.make_model(R_inner=R_inner, R_disc=R_disc)
        #       cube = model.get_cube(vchan_data, vel2d, int2d, linew2d, lineb2d, tb = {'nu': restfreq, 'beam': model.beam_info, 'full': False/True})
        #       cube = model.get_cube(vchan_data, vel2d, int2d, linew2d, lineb2d, make_convolve=False)
        #Cube.__init__(data, self.header, self.vchannels, beam=self.beam, filename=filename)    

    def _make_grid(self):
        dpix_rad = np.abs(self.header['CDELT1'])*u.Unit(self.header['CUNIT1']).to(u.radian)
        dpix_au = (self.dpc*np.tan(dpix_rad)).to(u.au)
        nx = self.header['NAXIS1']
        ny = self.header['NAXIS2']
        if nx!=ny:
            raise InputError((nx, ny), 'Number of spatial rows and columns in the datacube must be equal. '\
                             'Please clip the datacube using the datacube.clip(npix) method to produce a square cube.')
        
        xsky = ysky = ((nx-1) * dpix_au/2.0) # Sky maximum extent
        # dpix_au*nx = np.abs(-xsky - dpix_au/2) + (xsky + dpix_au/2) --> xsky is referred
        #  to the centre of the left/rightmost pixel. To recover the full extent of the sky,
        #   which should be equal to dpix_au*nx, one has to add twice half the pixel size to
        #    account for the total extent of the border pixels.
        grid = dgrid(xsky, nx) #Transforms xsky from au to metres and computes Cartesian grid
        self.skygrid = grid

        # The cell size of the discgrid is the same as that of the skygrid.
        #  The extent of the discgrid is the closest posible to the input Rmax using
        #   the aforementioned cell size.

        # discgrid should always exists, in general npix_disc != npix_sky,
        #  if npix_disc<npix_sky it's useful to use discgrid to save computing time,
        #   if npix_disc>npix_sky discgrid is needed for the emission to appear smooth
        #    instead of having sharp square boundaries.

        nx_disc = nx + int(np.round(2*self.Rmax.to(u.au)/dpix_au - nx))
        nx_disc += nx_disc%2 # making it even
        xdisc = (nx_disc-1)*dpix_au/2.0
        self.discgrid = dgrid(xdisc, nx_disc) #Transforms xdisc from au to metres and computes Cartesian grid
        
        if self.write_extent:
            log_grid = dict(nx=nx, ny=ny, xsky=xsky.value, xdisc=xdisc.value, cellsize=dpix_au.value, unit='au')
            with open("grid_extent.json", "w") as outfile:
                json.dump(log_grid, outfile, indent=4, sort_keys=False)



from discminer.mining_control import _mining_moment_offset
from discminer.mining_utils import (get_2d_plot_decorators,
                                    get_noise_mask,
                                    load_moments,
                                    load_disc_grid,
                                    init_data_and_model,
                                    mark_planet_location,
                                    overlay_continuum,
                                    show_output)

from discminer.rail import Contours
from discminer.plottools import (make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 get_cmap_from_color,
                                 use_discminer_style)

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from astropy import units as u

import json
import warnings

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_moment_offset(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']

rings = custom['rings']
gaps = custom['gaps']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)
    
file_data = meta['file_data']
tag = meta['tag']

dpc = meta['dpc']*u.pc
Rmax_frac = 1.1 #Max model radius, 10% larger than disc Rout
Rmax = Rmax_frac*Rout*u.au

#********************
#LOAD DATA AND GRID
#********************
#datacube = Data(file_data, dpc) # Read data and convert to Cube object
datacube, model = init_data_and_model(Rmin=0, Rmax=Rmax)
noise_mean, mask = get_noise_mask(datacube)

with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

R, phi, z = load_disc_grid()
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

#**************************
#MAKE PLOT + ZOOM-IN PANEL
fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(15,8))
ax_cbar0 = fig.add_axes([0.13, 0.09, 0.77, 0.05])

kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=levels_im)
kwargs_cc = dict(colors='k', linestyles='-', extent=extent, levels=levels_cc, linewidths=0.4)
kwargs_cbar = dict(orientation='horizontal', pad=0.03, shrink=0.95, aspect=15)
zoomcolor = '0.3'
zoomwidth = args.zoom_size

def make_plot(ax, xlim=xlim, color='k', labelcolor='k'):
    im = ax.contourf(moment_data, extend='both', **kwargs_im)
    if args.surface!='lower' and args.show_contours:
        cc = ax.contour(moment_data, **kwargs_cc)
    make_up_ax(ax, xlims=(-xlim, xlim), ylims=(-xlim, xlim), labelsize=17, color=color, labelcolor=labelcolor)
    return im

for i,axi in enumerate(ax):
    if i==1:
        im = make_plot(axi, xlim=zoomwidth, color=zoomcolor, labelcolor=zoomcolor)
        axi.scatter(best['orientation']['xc'], best['orientation']['yc'],
                    ec='k', fc='w', marker='X', lw=0.5+i, s=60*(i+1), zorder=20)                
    else:
        im = make_plot(axi, xlim=xlim)
        
    mod_major_ticks(axi, axis='both', nbins=8)
    datacube.plot_beam(axi, fc='0.8')
    axi.set_aspect(1)
    #model.disc_axes(axi)
    
for axi in ax[1:]:
    axi.tick_params(labelleft=False)
    for side in ['top','bottom','left','right']:
        axi.spines[side].set_linewidth(4.0)
        axi.spines[side].set_color(zoomcolor)
        axi.spines[side].set_linestyle((0, (1,1.5)))
        axi.spines[side].set_capstyle('round')
    axi.grid(color='k', ls='--')
    
for i,axi in enumerate([ax[0]]):
    Contours.emission_surface(axi, R, phi, extent=extent,
                              R_lev=np.arange(25, Rmax_frac*Rout, 50)*u.au.to('m'),
                              which=mtags['surf'],
                              kwargs_R={'colors': '0.1', 'linewidths': 0.5},
                              kwargs_phi={'colors': '0.1', 'linewidths': 0.4}
    )
    #R_lev=np.linspace(0.1, 1.0, 10)*Rout*

model.make_disc_axes(ax[0], surface=args.surface)
    
patch = Rectangle([-zoomwidth]*2, 2*zoomwidth, 2*zoomwidth, edgecolor=zoomcolor, facecolor='none',
                  lw=2.0, ls=(0, (1,1.5)), capstyle='round')
ax[0].add_artist(patch)

cbar0 = plt.colorbar(im, cax=ax_cbar0, format='%.1f', **kwargs_cbar)
cbar0.ax.tick_params(labelsize=15) 
mod_nticks_cbars([cbar0], nbins=10)
cbar0.set_label(clabel, fontsize=16)
ax[0].set_ylabel('Offset [au]', fontsize=17)
ax[0].set_title(ctitle, pad=40, fontsize=19)
ax[1].set_title('Zoom-in', pad=40, fontsize=19, color=zoomcolor)

#SHOW CONTINUUM?        
if args.show_continuum in ['all', 'scattered']:
    try:
        import cmasher as cmr
        cmap = plt.get_cmap('cmr.watermelon')
        overlay_continuum(ax[1], parfile='parfile_scattered.json', coords='sky', cmap=cmap, vmax=0.8, surface=args.surface, extend='max')
    except FileNotFoundError as e:
        warnings.warn('Unable to load parfile for scattered light image...', Warning)

if args.show_continuum in ['all', 'band7']:
    try:
        cmap = get_cmap_from_color('k', lev=32, vmin=0.0, vmax=0.2)
        overlay_continuum(ax[1], parfile='parfile_band7.json', coords='sky', lev=3, contours=True, cmap=cmap, surface=args.surface, zorder=20)
    except FileNotFoundError as e:
        warnings.warn('Unable to load parfile for band7 image...', Warning)

#*************
#MARK PLANETS
#*************
kwargs_sc = dict(s=150, lw=2.0, edgecolors='k')
for axi in ax:
    mark_planet_location(axi, args, dpc=dpc, coords='sky', zfunc=model.z_upper_func, zpars=best['height_upper'], **best['orientation'], **kwargs_sc)

plt.savefig('moment+offset_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)
show_output(args)
